"""Calculations related to calendars and astronomical parameters."""
import numpy as np
import xarray as xr
import xarray.ufuncs as xrf

#: Variable long-names.
LONG_NAMES = {
    'julian_day': 'Julian Day',
    'ndays_since': 'Days since 2000-01-01 12:00:00',
    'mean_longitude': 'Mean Longitude',
    'elevation': 'Elevation', 'zenith': 'Zenith',
    'azimuth': 'Azimuth',
    'mean_anomaly': 'Mean Anomaly',
    'ecliptic_longitude': 'Ecliptic Longitude',
    'ecliptic_obliquity': 'Obliquity of Ecliptic',
    'right_ascension': 'Right Ascension',
    'declination': 'Declination',
    'utc_hour': 'Hour UTC',
    'greenwich_mean_sideral_time': 'Greenwich Mean Sideral Time',
    'local_mean_sideral_time': 'Local Mean Sideral Time',
    'hour_angle': 'Hour Angle',
    'sunset_hour_angle': 'Sunset Hour Angle',
    'daily_cos_zenith': 'Daily-Integrated Cosine of Zenith',
    'daily_mean_cos_zenith': 'Daily-Averaged Cosine of Zenith'
}


def _compute_if_none(fun):
    """Compute attribute with the same name as :py:obj:`fun`
    only if `None`."""

    def comp_fun(self):
        attr_name = '_' + fun.__name__

        # Compute variable only if None
        if getattr(self, attr_name) is None:
            # Compute variable
            var = fun(self)

            # Make sure it is a data array
            try:
                var = xr.DataArray(var, coords=[('time', self.time)])
            except ValueError:
                var = xr.DataArray(var)

            # Add variable to attribute
            setattr(self, attr_name, var)

            # Add names
            self._add_attributes(attr_name)

        return getattr(self, attr_name)

    return comp_fun


def _clip_angle(angle):
    """Clip angle so that: math: `-\pi ^\\circ <= \\theta < \pi ^\\circ`.

    :param angle: Angle in radians.
    :type angle: (sequence of) :py:class:`float`

    :returns: Clipped angle.
    :rtype: (sequence of) :py:class:`float`
    """
    lim = np.pi
    return xrf.fabs(xrf.fmod(angle + lim, 2 * lim)) - lim


class SolarCalendarComputer(object):
    """Class to compute solar calendar variables.
    When getting an attribute, the variable
    corresponding to it is automatically computed. New computations are
    only performed if the needed attribute value is `None`, i.e. if it
    has not previously been computed.
    To perform new computations, start from a new object,
    or call :py:meth:`clear`.

    .. note::
      * The Almanac approximation is used with ecliptic coordinates,
        as in Michalsky (1988).
      * The *solar elevation* is the angle between the horizontal and the
        line to the sun, that is, the complement of the *zenith angle*.
      * The *solar azimuth* is the angular displacement from south of the
        projection of beam radiation on the horizontal plane. Displacements
        east of south are negative and west of south are positive.
      * The *right ascension* is the angular distance measured eastward
        along the celestial equator from the Sun at the March equinox
        to the hour circle of the point above the earth in question.
      * The *declination* is the angular position of the sun at solar noon
        (i.e., when the sun is on the local meridian)
        with respect to the plane of the equator, north positive;
        :math:`−23.45^\circ \le \delta \le 23.45^\circ`.
      * The *hour angle* is the angular displacement of the sun east or west
        of the local meridian due to rotation of the earth on its axis at
        15 degrees per hour; morning negative, afternoon positive.
      * The *sunset hour angle* is the hour angle when elevation is zero.
        See Duffie and Beckman(2013).

    .. warning::
      The Almanac's approximation covers the (1950-2050) period
      and may need to be modified for longer periods.

    .. seealso::
      * Michalsky, J.J., 1988. The Astronomical Almanac's Algorithm format
        Approximate Solar Position (1950, 2050). *Solar Energy* 40, 227-235.
      * Duffie, J.A., Beckman, W.A., 2013.
      * *Solar Energy and Thermal Processes*, fourth ed. Wiley, Hoboken, NJ.

    """

    def __init__(self, time, lat, lon, angles_in_degrees=False):
        """Initialize solar angles computer.

        :param time: Time.
        :param lat: Latitude.
        :param lon: Longitude.
        :param angles_in_degrees: Whether angles are given in degrees instead
          of radians and should be returned in degrees by :py:meth:`get_angle`.
        :type time: (sequence of) datetime.
        :type lat: (sequence of) :py:class:`float`.
        :type lon: (sequence of) :py:class:`float`.
        :type angles_in_degrees: bool

        .. warning::

          * All attribute variables (e.g. :py:attr:`hour_angle`)
            are given in radians, between :py:math:`-\pi` and :py:math:`\pi`,
            and all computations are performed
            in radians. To get angles in the same units as those
            provided to constructor, call :py:meth:`get_angle`.
          * 
        """

        # Manage degrees-radians conversion
        self.angles_in_degrees = angles_in_degrees
        self.fact = np.deg2rad(1.) if self.angles_in_degrees else 1.
        self.inv_fact = 1. if self.angles_in_degrees else np.deg2rad(1.)
        self.angle_units = 'degrees' if self.angles_in_degrees else 'radians'

        # Convert to radians if needed
        self.time = time
        self.lat = lat * self.fact
        self.lon = lon * self.fact

        # Initialize all variables to be computed to None
        self.clear()

    def clear(self):
        """Set all computed attribute variables to `None`."""
        for variable_name in LONG_NAMES:
            setattr(self, '_' + variable_name, None)

    def _add_attributes(self, variable_name):
        """Add `long_name` and `units` attributes to data array.

        :param variable_name: Variable name.
        :type variable_name: str
        """
        # Get variable from variable name
        var = getattr(self, variable_name)

        # Try to add attributes
        try:
            variable_name = (variable_name[1:] if variable_name[0] == '_' else
                             variable_name)
            var.attrs['long_name'] = LONG_NAMES[variable_name]
            var.attrs['units'] = self.angle_units
        except AttributeError:
            pass

    def get_angle(self, angle_name):
        """Get angle, converted in degrees if needed.

        :param angle_name: Angle name.
        :type angle_name: str

        .. warning:: No exception is raised if a non-angle variable name,
          e.g. :py:attr:`julian_day` is given, leading to return this variable
          after a potential spurious angle conversion. Instead, such variables
          should directly be accessed as attribute.
        """
        return _clip_angle(getattr(self, angle_name)) / self.fact

    @property
    @_compute_if_none
    def julian_day(self):
        """Get Julian day as data array."""
        return self.time.to_julian_date()

    @property
    @_compute_if_none
    def ndays_since(self):
        """Get number of days since 2000-01-01 12:00:00."""
        return self.julian_day - 2451545.0

    @property
    @_compute_if_none
    def mean_longitude(self):
        """Get mean longitude :py:math:`-\pi <= mean_lon <= \pi`."""
        return 4.89495 + 0.01720279 * self.ndays_since

    @property
    @_compute_if_none
    def elevation(self):
        """Compute solar elevation from Julian day, East longitude and
        latitude. Angular extent of sun neglected.
        """
        # Elevation (radians)
        return xrf.arcsin(
            xrf.sin(self.declination) * xrf.sin(self.lat) +
            xrf.cos(self.declination) * xrf.cos(self.lat) *
            xrf.cos(self.hour_angle))

    @property
    @_compute_if_none
    def zenith(self):
        """Get zenith from elevation angle."""
        return np.pi/2 - self.elevation

    @property
    @_compute_if_none
    def azimuth(self):
        """Compute solar azimuth from Julian day, East longitude and latitude.
        Angular extent of sun neglected.
        """
        # Azimuth (radians)
        # Numerical errors may lead to absolute sine values slighly larger
        # than 1. In this case, angle is set to -pi/2, as should be.
        var = -np.pi/2 * xr.ones_like(self.elevation)
        val = (-xrf.cos(self.declination) * xrf.sin(self.hour_angle)
               / xrf.cos(self.elevation))
        xrf.arcsin(val, out=var.values, where=(xrf.fabs(val) < 1.))

        # Assign correct quadrant to azimuth (check discontinuity)
        elc = xrf.arcsin(xrf.sin(self.declination) / xrf.sin(self.lat))
        var = xr.where(self.elevation >= elc, np.pi - var, var)

        return var

    @property
    @_compute_if_none
    def mean_anomaly(self):
        """Get mean anomaly."""
        return 6.24004 + 0.01720197 * self.ndays_since

    @property
    @_compute_if_none
    def ecliptic_longitude(self):
        """Get ecliptic longitude."""
        return self.mean_longitude + 0.03342 * xrf.sin(
            self.mean_anomaly) + 0.00035 * xrf.sin(2 * self.mean_anomaly)

    @property
    @_compute_if_none
    def ecliptic_obliquity(self):
        """Get ecliptic obliquity."""
        return 0.40908 - 7.e-9 * self.ndays_since

    @property
    @_compute_if_none
    def right_ascension(self):
        """Get right ascension."""
        return xrf.arctan2(
            xrf.cos(self.ecliptic_obliquity) *
            xrf.sin(self.ecliptic_longitude),
            xrf.cos(self.ecliptic_longitude))

    @property
    @_compute_if_none
    def declination(self):
        """Get declination."""
        return xrf.arcsin(
            xrf.sin(self.ecliptic_obliquity) *
            xrf.sin(self.ecliptic_longitude))

    @property
    @_compute_if_none
    def utc_hour(self):
        """Get UTC hour in radians."""
        # A decimal part of the Julian date equal to zero
        # corresponds to noon (12h)
        djd = self.julian_day.astype(int)
        return (self.julian_day - djd) * 2*np.pi + np.pi

    @property
    @_compute_if_none
    def greenwich_mean_sideral_time(self):
        """Get Greenwich mean sideral time."""
        return 1.753369 + 0.0172027917 * self.ndays_since + self.utc_hour

    @property
    @_compute_if_none
    def local_mean_sideral_time(self):
        """Get local mean sideral time."""
        return self.greenwich_mean_sideral_time + self.lon

    @property
    @_compute_if_none
    def hour_angle(self):
        """Get hour angle in radians."""
        return self.local_mean_sideral_time - self.right_ascension

    @property
    @_compute_if_none
    def sunset_hour_angle(self):
        """Compute sunset hour angle from latitude and declination."""
        # Compute sunset hour angle
        return xrf.arccos(
            -xrf.tan(self.lat) * xrf.tan(self.declination))

    @property
    @_compute_if_none
    def daily_cos_zenith(self):
        """Get daily-integrated cosine of zenith angle."""
        return ((
            self.sunset_hour_angle * xrf.sin(self.declination)
            * xrf.sin(self.lat) + xrf.cos(self.declination) * xrf.cos(self.lat)
            * xrf.sin(self.sunset_hour_angle))
            * 2 * (12. / np.pi))

    @property
    @_compute_if_none
    def daily_mean_cos_zenith(self):
        """Get daily-averaged cosine of zenith angle."""
        return xr.where(
            xrf.fabs(self.sunset_hour_angle) > 1.e-8,
            self.daily_cos_zenith / (2 * (12. / np.pi))
            / self.sunset_hour_angle, 0.)



import os
import requests
from pkg_resources import resource_filename
import pandas as pd
import xarray as xr
from ..data_source import DataSourceLoaderBase
from ..component import (
    DataSourceWithComponentsMixin, parse_variable_component_args,
    download_to_compute_capacity_factor, compute_capacity_factor,
    finalize_dataset)


class DataSource(DataSourceLoaderBase, DataSourceWithComponentsMixin):
    #: Default source name.
    DEFAULT_SRC_NAME = 'energidataservice_dk'

    #: Area: `'Denmark'`
    AREA = 'Denmark'

    def __init__(self, med, name=None, cfg=None, **kwargs):
        """Naming constructor.

        :param med: Mediator.
        :param name: Data source name.
        :param cfg: Data source configuration. Default is `None`.
        :type med: :py:class:`.mediator.Mediator`
        :type name: str
        :type cfg: mapping
        """
        name = name or self.DEFAULT_SRC_NAME
        super(DataSource, self).__init__(med, name, cfg=cfg, **kwargs)

    @parse_variable_component_args
    @download_to_compute_capacity_factor
    def download(self, variable_component_names, **kwargs):
        """Download data.

        :param variable_component_names: Names of components to download per
          variable.
        :type variable_component_names: mapping from :py:class:`str`
          to :py:class:`set` of :py:class:`str`

            :returns: Names of downloaded components per variable.
        :rtype: mapping from :py:class:`str` to :py:class:`set` of
          :py:class:`str`

        .. note:: The data is downloaded from the
          `energidataservice.dk API <https://api.energidataservice.dk/>`
        """
        # Get municipality codes
        code_region = self._get_code_region(**kwargs)

        # Loop over variables
        for variable_name, component_names in variable_component_names.items():
            if not self.cfg.get('no_verbose'):
                self.log.info('- {}:'.format(variable_name))

            # Manage wind component as onshore plus offshore
            component_name = 'wind'
            component_to_download_names = component_names.copy()
            if component_name in component_names:
                # Make sure that generation and capacity are downloaded
                component_to_download_names.update(
                    ['wind-onshore', 'wind-offshore'])

                # Remove capacity factor from components to download
                component_to_download_names.discard(component_name)

            # Loop over components
            for component_name in component_to_download_names:
                # Get URL
                url = self._get_url(variable_name, component_name, **kwargs)
                filepath = self._get_download_filepath(
                    variable_name, component_name, **kwargs)
                if not self.cfg.get('no_verbose'):
                    self.log.info('-- {} from {} to {}'.format(
                        component_name, url, filepath))

                # Request and raise exception if needed
                response = requests.get(url)
                response.raise_for_status()
                records = response.json()['result']['records']

                # Convert to data frame with time as index
                df = pd.DataFrame(records).set_index(self.cfg['frequency'])
                df.index.name = 'time'

                # Sort data frame by index
                df = df.iloc[df.index.argsort()]

                # Remove 0 and 1 codes
                df = df.loc[(df['MunicipalityNo'].astype(int) != 0) &
                            (df['MunicipalityNo'].astype(int) != 1)]

                # Add municipality names
                df['region'] = [code_region.get(int(c))
                                for c in df['MunicipalityNo']]

                # Write data frame
                df.to_csv(filepath)

        # Return names of downloaded variables
        return variable_component_names

    @parse_variable_component_args
    @compute_capacity_factor
    @finalize_dataset
    def load(self, variable_component_names, **kwargs):
        """Load data.

        :param variable_component_names: Names of components to load per
          variable.
        :type variable_component_names: mapping from :py:class:`str`
          to collection

        :returns: Time series for each variable and component.
        :rtype: :py:class:`xarray.Dataset()`
        """
        # Get region names
        _, src_region_place = self.med.geo_src.get_place_regions_for_source(
            self.name, self.AREA, **kwargs)

        # Loop over variables
        ds = xr.Dataset()
        for variable_name, component_names in variable_component_names.items():
            if not self.cfg.get('no_verbose'):
                self.log.info('- {}:'.format(variable_name))

            # Manage wind component as onshore plus offshore
            component_name = 'wind'
            component_to_load_names = component_names.copy()
            if component_name in component_names:
                # Make sure that generation and capacity are loaded
                component_to_load_names.update(
                    ['wind-onshore', 'wind-offshore'])

                # Remove capacity factor from components to load
                component_to_load_names.discard(component_name)

            # Loop over components
            da = None
            for component_name in component_to_load_names:
                # Get filepath
                filepath = self._get_download_filepath(
                    variable_name, component_name, **kwargs)
                if not self.cfg.get('no_verbose'):
                    self.log.info('-- {} from {}'.format(
                        component_name, filepath))

                # Read data frame
                df = pd.read_csv(filepath, index_col=0, header=0)

                # Aggregate in zones/regions
                df_place = self._aggregate_places(
                    df, src_region_place, component_name, variable_name,
                    **kwargs)

                # Convert to data array
                da_place = xr.DataArray(
                    df_place, dims=('time', 'region')).expand_dims(
                        'component').assign_coords(component=[component_name])

                # Add place data
                da = (da_place if da is None else
                      xr.concat([da, da_place], dim='component'))

                # Make sure that time index is datetime
                da['time'] = pd.DatetimeIndex(da.indexes['time'])

            # Add total wind component if needed
            component_name = 'wind'
            if component_name in component_names:
                da_wind = (da.sel(component='wind-onshore')
                           + da.sel(component='wind-offshore')).expand_dims(
                               'component').assign_coords(
                                   component=[component_name])
                da = xr.concat([da, da_wind], dim='component')

            # Select requested components
            da = da.sel(component=list(component_names))

            # Add variable to dataset
            ds[variable_name] = da

        return ds

    def _aggregate_places(self, df, src_region_place, component_name,
                          variable_name, **kwargs):
        """Aggregate data frame in places.

        :param df: Data frame to aggregate.
        :param src_region_place: Region-place mapping for data source.
        :param component_name: Component name.
        :param variable_name: Variable name.
        :type df: :py:class:`pandas.DataFrame`
        :type src_region_place: mapping
        :type component_name: str
        :type variable_name: str

        :returns: Aggregated data frame.
        :rtype: :py:class:`pandas.DataFrame`
        """
        df['place'] = [src_region_place[region_name]
                       for region_name in df['region']]
        src_component_name = '{}{}'.format(
            self.cfg['component_names'][component_name],
            self.cfg['variable_names_in_components'][variable_name])
        df = df[[src_component_name, 'place']]
        df_place = df.groupby(['time', 'place']).sum().unstack()[
            (src_component_name,)]
        df_place.columns.name = 'region'

        return df_place

    def _get_capacity_factor(self, ds, **kwargs):
        """Get capacity factor from generation and capacity.

        :param ds: Dataset containing generation and capacity.
        :type ds: mapping

        :returns: Capacity factor.
        :rtype: :py:class:`xarray.DataArray`
        """
        da_gen = ds['generation']

        # Get number of days in month
        time = da_gen.indexes['time']
        start = pd.Timestamp(time[0])
        end = pd.Timestamp(time[-1]) + pd.Timedelta('32 days')
        time_plus = pd.date_range(start=start, end=end, freq='MS')
        ndays = (time_plus[1:] - time_plus[:-1]).days
        da_ndays = xr.DataArray(ndays, coords=[('time', time)])
        da_nhours = da_ndays * 24

        # Get capacity factor
        da = da_gen / (ds['capacity'] * da_nhours)

        return da

    def _get_code_region(self, **kwargs):
        """Get regions corresponding to municipality codes.

        :returns: Code-to-region mapping.
        :rtype: mapping from :py:class:`str` to :py:class:`str`.
        """
        src_name = 'dst'
        filename = 'municipality_codes_regions.csv'
        resource_name = '../data/{}/{}/{}'.format(
            self.AREA, src_name, filename)

        # Read
        df = pd.read_csv(resource_filename(__name__, resource_name),
                         sep=';')

        code_regions = dict(zip(df['CODE'].astype(int), df['REGION']))

        return code_regions

    def _get_download_filepath(self, variable_name, component_name, **kwargs):
        """Write downloaded data for variable and component.

        :param variable_name: Name of data variable.
        :param component_name: Name of data component.
        :type variable_name: str
        :type component_name: str

        :returns: Filepath.
        :rtype: str
        """
        src_dir = self.med.cfg.get_external_data_directory(self)
        filename = '{}_{}_{}{}.csv'.format(
            self.name, component_name, variable_name,
            self.get_postfix(**kwargs))
        filepath = os.path.join(src_dir, filename)

        return filepath

    def _get_url(self, variable_name, component_name, **kwargs):
        """Get URL for query.

        :param variable_name: Name of variable to query.
        :param component_name: Name of component to query.
        :type variable_name: str
        :type component_name: str

        :returns: URL.
        :rtype: str
        """
        src_variable_name = self.cfg['variable_names'][variable_name]
        src_component_name = '{}{}'.format(
            self.cfg['component_names'][component_name],
            self.cfg['variable_names_in_components'][variable_name])
        res_label = 'res'

        # Component selection
        comp_query = '{}."{}"'.format(res_label, src_component_name)

        # Municipality query
        place_query = '{}."{}"'.format(res_label, 'MunicipalityNo')

        # Time query and selection
        time_query = '{}."{}"'.format(res_label, self.cfg['frequency'])
        time_selection = (
            'WHERE {}."{}" >= timestamp\'{}\' '
            'AND {}."{}" < timestamp\'{}\''.format(
                res_label, self.cfg['frequency'], self.cfg['start_date'],
                res_label, self.cfg['frequency'], self.cfg['end_date']))

        # SQL query
        sql_query = 'SELECT {}, {}, {} FROM "{}" as {} {}'.format(
            time_query, place_query, comp_query, src_variable_name, res_label,
            time_selection)

        # URL
        url = '{}/datastore_search_sql?sql={}'.format(
            self.cfg['host'], sql_query)

        return url

    def get_postfix(self, **kwargs):
        """Get data-source standard postfix.

        :returns: Postfix.
        :rtype: str
        """
        postfix = '_{}-{}'.format(
            self.cfg['start_date'], self.cfg['end_date'])

        return postfix



"""Bias corrector and predictor."""
import operator
from numpy import set_printoptions
import pandas as pd
from sklearn.exceptions import NotFittedError
from ..actuator_base import EstimatorBase
from ..utils.learn import parse_fit_data, select_data


class Actuator(EstimatorBase):
    """Bias corrector and predictor."""

    def __init__(self, result_mng, name, cfg=None, **kwargs):
        """Naming constructor.

        :param result_mng: Result manager of variable to estimate.
        :param name: Actuator name.
        :param cfg: Actuator configuration. Default is `None`.
        :type result_mng: :py:class:`ResultManager`
        :type name: str
        :type cfg: mapping
        """
        super(Actuator, self).__init__(
            result_mng=result_mng, name=name, cfg=cfg,
            **kwargs)

        #: Operator choice.
        self._operator_choice = self.cfg.get('operator')
        self._operator_choice = ('scale' if self._operator_choice is None
                                 else self._operator_choice)

        #: Bias-correction operator.
        #: Set to :py:meth:`operator.mul`, or :py:meth:`operator.add`,
        #: if :py:attr:`_operator_choice` is `'scale'`, or `'shift'`,
        #: respectively.
        self._operator = None

        #: Inverse of bias-correction operator
        #: Set to :py:meth:`operator.truediv`, or :py:meth:`operator.sub`,
        #: if :py:attr:`_operator_choice` is `'scale'`, or `'shift'`,
        #: respectively.
        self._operator_inv = None

        if self._operator_choice == 'scale':
            self._operator = operator.mul
            self._operator_inv = operator.truediv
        elif self._operator_choice == 'shift':
            self._operator = operator.add
            self._operator_inv = operator.sub

    def fit(self, data_src_in, data_src_out, **kwargs):
        """Get bias corrector for component as a factor to multiply the
        input data with and such that the training input data
        multiplied by the bias corrector has the same mean
        as the training output data over the same period.

        :param data_src_in: Training set input data source.
        :param data_src_out: Training set output data source.
        :type data_src_in: :py:class:`.data_source.DataSourceBase`
        :type data_src_out: :py:class:`.data_source.DataSourceBase`

        .. warning:: If the `intersection_only_if_possible` entry of the
          container configuration is `True`, an attempt is made to
          compute the bias on a common time-slice. The time indices of
          the two data sources should, however, be compatible.

        .. seealso:: :py:func:`_intersection_time_slice`
        """
        # Get input and output data arrays
        da_in, da_out = parse_fit_data(
            data_src_in, data_src_out, self.result_mng,
            subsample_freq=self.cfg.get('subsample_freq'),
            select_period=True, time_slice=self.cfg.get('time_slice'),
            **kwargs)

        # Check if selected period intersects that of input period
        time_slice_in = slice(*da_in.indexes['time'][[0, -1]])
        time_slice_out = slice(*da_out.indexes['time'][[0, -1]])
        if self.cfg.get('intersection_only_if_possible'):
            inter_slice = _intersection_time_slice(da_in, da_out)
            if inter_slice is not None:
                time_slice_in = time_slice_out = inter_slice

        # Get bias corrector
        if not self.cfg.get('no_verbose'):
            self.log.info(
                'Computing {} {} bias corrector from {} to {} (input) '
                'and from {} to {} (output)'.format(
                    self.result_mng.name, self.component_mng.name,
                    time_slice_in.start, time_slice_in.stop,
                    time_slice_out.start, time_slice_out.stop))
        mean_in = da_in.sel(time=time_slice_in).mean('time')
        mean_out = da_out.sel(time=time_slice_out).mean('time')
        bc = self._operator_inv(mean_out, mean_in)

        # Convert coordinates datatypes from object to str if needed
        for dim in bc.dims:
            if bc[dim].dtype == object:
                bc[dim] = bc[dim].astype(str)
        self.coef = bc

        set_printoptions(precision=3)
        if not self.cfg.get('no_verbose'):
            self.log.info('Computed average ({} / {}):'.format(
                time_slice_in.start, time_slice_in.stop))
            self.log.info(mean_in.values)
            self.log.info('Observed average ({} / {}):'.format(
                time_slice_out.start, time_slice_out.stop))
            self.log.info(mean_out.values)

    def predict(self, data_src_in, **kwargs):
        """Apply bias corrector for component by multipling input data.

        :param data_src_in: Input data source.
        :type data_src_in: :py:class:`..data_source.DataSourceBase`

        :returns: Bias corrected dataset.
        :rtype: dict
        """
        ds = {}
        # Verify that bias corrector has been fitted
        if self.coef is None:
            raise NotFittedError('This bias corrector instance '
                                 'must be fitted before prediction.')

        # Select result, copy and ensure dimensions order
        da_in = select_data(data_src_in, self.result_mng, **kwargs)

        # Copy input dataset
        da_pred = da_in.copy(deep=True)

        # Reorder regions to comply
        coef_comp = (self.coef.loc[{'region': da_in.indexes['region']}]
                     if 'region' in self.coef.coords else self.coef.copy())

        # Select component if needed
        try:
            coef_comp = coef_comp.sel(
                component=self.component_mng.component_name)
        except ValueError:
            pass

        # Apply bias corrector, with input data
        da_pred = self._operator(da_pred, coef_comp)

        # Add bias-corrected output variable to dataset
        ds[self.result_mng.result_name] = da_pred

        return ds

    def _select_period_from_array(self, da, data_type):
        """Select full years period from an array.
        Try to get first and last dates from configuration.
        Otherwise, select the first full years.

        :param da: Array.
        :param data_type: `'in'` or '`out`' data type for which
          to get dates in configuration.
        :type da: :py:class:`xarray.DataArray`
        :type data_type: str

        :returns: Tuple containing the selected array and time slice.
        :rtype: :py:class:`tuple` of :py:class:`xarray.DataArray`
          and :py:class:`slice`
        """
        time_slice = self.cfg.get('{}_time_slice'.format(data_type))
        if time_slice is None:
            # Get time index
            time = da.indexes['time']

            # First date is given or first available date
            first_date = pd.Timestamp(
                self.cfg.get('first_date_{}'.format(data_type)) or
                time[0])

            # Last date is given or last date which is a multiple of one
            # year past first date
            last_date = pd.Timestamp(
                self.cfg.get('last_date_{}'.format(data_type)) or
                _last_full_years_date(time, first_date))
        else:
            first_date, last_date = time_slice

        # Slice
        time_slice = slice(first_date, last_date)
        da = da.sel(time=time_slice)

        return da, time_slice

    def get_estimator_postfix(self, **kwargs):
        """Get bias-corrector postfix.

        returns: Postfix.
        rtype: str
        """
        return '{}_nobias'.format(
            super(Actuator, self).get_estimator_postfix(**kwargs))


def _last_full_years_date(time, first_date):
    """Get last date in date-time index which is a multiple of a year
    away from start date.

    :param time: Date-time index.
    :param first_date: First date.
    :type time: :py:class:`pandas.DatetimeIndex`
    :type first_date: :py:class:`datetime.date`

    :returns: Last date.
    :rtype: :py:class:`datetime.date`
    """
    last_date = pd.date_range(time[0], time[-1], freq='A-DEC')[-1]

    # Make sure last samples are present
    freq = time.inferred_freq
    if freq is None:
        if (time[1] - time[0]).seconds == 3600:
            freq = 'H'
    if freq == 'H':
        last_date += pd.Timedelta('23H')

    return last_date


def _intersection_time_slice(da_in, da_out):
    """Get full years intersection between two arrays, if possible.

    :param da_in: First array.
    :param da_out: Second array.
    :type da_in: :py:class:`xarray.DataArray`
    :type da_out: :py:class:`xarray.DataArray`

    :returns: Intersection time slice.
    :rtype: slice

    .. warning:: The time index of the input and output arrays
      must be compatible. For instance, daily-sampled indices
      but with different hours won't intersect.
    """
    # Match frequencies
    t_out = da_out.indexes['time']
    freq_out = t_out.inferred_freq
    t_in = da_in.indexes['time']
    freq_in = t_in.inferred_freq
    if freq_out in ['A-DEC', 'Y']:
        # Make index of yearly mean start 1st of January
        da_out['time'] = pd.DatetimeIndex(
            ['{}-01-01'.format(t.year) for t in t_out])
        if freq_in == 'H':
            start = '{}-01-01 00:00:00'.format(t_out[0].year)
            end = '{}-12-31 23:00:00'.format(t_out[-1].year)
            t_out_hour = pd.date_range(start, end, freq=freq_in)
            da_out = da_out.reindex(time=t_out_hour, method='ffill')

    t_inter = da_out.indexes['time'].intersection(
        da_in.indexes['time'])
    if len(t_inter) > 0:
        first_date_inter = t_inter[0]
        last_date_inter = _last_full_years_date(
            t_inter, first_date_inter)
        if ((last_date_inter - first_date_inter)
                >= pd.Timedelta(1, unit='Y')):
            # If there are at least a year of common data,
            # use common data only
            time_slice = slice(first_date_inter, last_date_inter)

            return time_slice



"""Modifier to resample."""
from ..actuator_base import ExtractorBase


class Actuator(ExtractorBase):

    def __init__(self, result_mng, name, cfg=None, **kwargs):
        """Naming constructor.

        :param result_mng: Result manager of variable to estimate.
        :param name: Actuator name.
        :param cfg: Actuator configuration. Default is `None`.
        :type result_mng: :py:class:`ResultManager`
        :type name: str
        :type cfg: mapping
        """
        super(Actuator, self).__init__(
            result_mng=result_mng, name=name, cfg=cfg, **kwargs)

    def transform(self, ds=None, data_src=None, **kwargs):
        """Apply modifier to data source.

        :param ds: Dataset to transform.
        :type ds: :py:class:`xarray.Dataset`

        :returns: Modified dataset.
        :rtype: Same as :py:obj:`data_src.data`
        """
        coord = list(self.cfg['indexer'])[0]

        ds_new = ds.copy()
        for da_name, da in ds.items():
            # Resample array
            gp = da.resample(self.cfg['indexer'])

            # Apply reduction
            ds_new[da_name] = getattr(gp, self.cfg['apply'])(coord)

        return ds_new

    def get_extractor_postfix(self, **kwargs):
        """Get extractor postfix.

        returns: Postfix.
        rtype: str
        """
        postfix = '_{}{}{}'.format(
            *list(self.cfg['indexer'].items())[0], self.cfg['apply'])

        return '{}{}'.format(
            super(Actuator, self).get_extractor_postfix(**kwargs), postfix)



"""ISTAT API."""
import os
from ..geo import DefaultMaskAPI, DefaultMaskMaker


class DataSource(DefaultMaskAPI, DefaultMaskMaker):
    def __init__(self, med, name=None, cfg=None, **kwargs):
        """Naming constructor.

        :param med: Mediator. Default is `None`.
        :param name: Data source name.
        :param cfg: Data source configuration.
        :type med: :py:class:`.mediator.Mediator`
        :type name: str
        :type cfg: mapping
        """
        name = name or 'istat'
        super(DataSource, self).__init__(med, name, cfg=cfg, **kwargs)

    def get_url_filename(self, *kwargs):
        """ Get URL and filename of geographical data.

        :returns: URL pointing to geographical data and
          the name of shapefile.
        :rtype: tuple of str
        """
        postfix = self.cfg['date'] + self.cfg['resolution']
        dir0 = 'Limiti' + postfix
        url = '{}/{}/{}.zip'.format(self.cfg['host'], self.cfg['path'], dir0)

        return url

    def get_filename(self, *args, **kwargs):
        """Get downloaded-file name.

        :returns: Filename.
        :rtype: str
        """
        postfix = self.cfg['date'] + self.cfg['resolution']
        dir0 = 'Limiti' + postfix
        dir1 = 'Reg' + postfix
        filename = os.path.join(dir0, dir1, dir1 + '.shp')

        return filename



"""Actuator base definitions."""
import os
import sys
from collections import OrderedDict
from abc import ABC, abstractmethod
import pickle
import xarray as xr
from .container import Container, ensure_collection
from .config import import_module_from_cfg


class ActuatorBase(Container):
    """Actuator base class."""

    def __init__(self, result_mng, name, cfg=None, task_names=set(),
                 **kwargs):
        """Actuator constructor.

        :param result_mng: Result manager of variable to estimate.
        :param name: Actuator name.
        :param cfg: Estimator configuration. Default is `None`.
        :param task_names: Names of potential tasks for container to perform.
          Default is `set()`.
        :type result_mng: :py:class:`ResultManager`
        :type name: str
        :type cfg: mapping
        :type task_names: set
        """
        #: Data sources.
        self.data_sources = None

        # Attach component and output variable to actuator
        #: Result manager of variable to act on.
        self.result_mng = result_mng
        #: Result component.
        self.component_mng = self.result_mng.component_mng

        # Initialize as container with output variable as parent
        kwargs.update({
            'component_mng': self.component_mng, 'name': name,
            'parent': result_mng, 'result_mng': result_mng, 'cfg': cfg,
            'med': self.component_mng.med, 'task_names': task_names})
        super(ActuatorBase, self).__init__(**kwargs)

        # Add data sources
        if 'data' in self.cfg:
            self.data_sources = OrderedDict()
            self.add_data_sources(**kwargs)

    def add_data_sources(self, cfg=None, **kwargs):
        """Add data sources from configuration.

        :param cfg: Configuration from which to get data source.
          Default is `None`, in which case the actuator configuration is
          used.
        :type cfg: mapping
        """
        cfg = cfg or self.cfg
        for stage, src_vars in cfg['data'].items():
            # Add data source to mediator and result manager
            self.data_sources[stage] = self.med.add_data_source(
                src_vars, self.result_mng)

            # Add data sources as children
            self.update_children({self.data_sources[stage].name:
                                  self.data_sources[stage]})


class EstimatorBase(ActuatorBase, ABC):
    """Estimator abstract base class. Requires :py:meth:`fit` and
    :py:meth:`predict` methods to be implemented."""

    def __init__(self, result_mng, name, cfg=None, **kwargs):
        """Estimator constructor.

        :param result_mng: Result manager of variable to estimate.
        :param name: Actuator name.
        :param cfg: Estimator configuration. Default is `None`.
        :type result_mng: :py:class:`ResultManager`
        :type name: str
        :type cfg: mapping
        """
        #: Coefficients to be fitted
        self.coef = None

        # Try updating task names in keyword arguments if possible
        task_names = kwargs.get('task_names') or set()
        task_names.update({'fit', 'predict'})
        kwargs.update({
            'component_mng': result_mng.component_mng, 'name': name,
            'result_mng': result_mng,
            'cfg': cfg, 'med': result_mng.med, 'task_names': task_names})
        # Initialize as container
        super(EstimatorBase, self).__init__(**kwargs)

    @abstractmethod
    def fit(**kwargs):
        """Fit estimator abstract method."""
        raise NotImplementedError

    @abstractmethod
    def predict(**kwargs):
        """Predict with estimator abstract method."""
        raise NotImplementedError

    def get_estimator_postfix(self, **kwargs):
        """Default implementation: get an empty postfix string.

        :returns: Postfix.
        :rtype: str
        """
        return ''

    def get_fit_postfix(self, **kwargs):
        """Get fit postfix (component, feature, estimator).

        :returns: Postfix.
        :rtype: str
        """
        # Feature postfix
        feature_postfix = self.result_mng.feature[
            'fit'].get_data_postfix(**kwargs)

        # Output data postfix
        data_src = self.result_mng.extractor['output'].data_sources[
            'output']
        output_postfix = data_src.get_data_postfix(
            with_src_name=True, **kwargs)

        # Estimator postfix
        estimator_postfix = self.get_estimator_postfix(**kwargs)

        # Fit postfix
        fit_postfix = '{}{}{}'.format(
            feature_postfix, output_postfix, estimator_postfix)

        return fit_postfix

    def get_fit_path(self, makedirs=True, **kwargs):
        """Get fit filepath.

        :param makedirs: Make directories if needed. Default is `True`.
        :type makedirs: bool

        :returns: Filepath.
        :rtype: str
        """
        filename = '{}_estimator{}'.format(
            self.component_mng.name, self.get_fit_postfix(**kwargs))
        data_dir = self.component_mng.get_data_dir(makedirs=makedirs, **kwargs)
        filepath = os.path.join(data_dir, filename)

        return filepath

    def read(self, **kwargs):
        """Read estimator with pickle."""
        try:
            # Try to read as netcdf data-array
            filepath = '{}.nc'.format(self.get_fit_path(**kwargs))
            if not self.cfg.get('no_verbose'):
                self.log.info('Reading {} {} {} estimator from {}'.format(
                    self.component_mng.name, self.result_mng.name, self.name,
                    filepath))
            self.coef = xr.load_dataarray(filepath)
        except FileNotFoundError:
            # Read as pickle otherwise
            filepath = '{}.pickle'.format(self.get_fit_path(**kwargs))
            if not self.cfg.get('no_verbose'):
                self.log.info('Reading {} {} {} estimator from {}'.format(
                    self.component_mng.name, self.result_mng.name, self.name,
                    filepath))
            with open(filepath, 'rb') as f:
                self.coef = pickle.load(f)

    def write(self, **kwargs):
        """Write estimator with pickle."""
        try:
            # Try to write as netcdf data-array
            filepath = '{}.nc'.format(self.get_fit_path(**kwargs))
            if not self.cfg.get('no_verbose'):
                self.log.info('Writing {} {} {} estimator to {}'.format(
                    self.component_mng.name, self.result_mng.name, self.name,
                    filepath))
            self.coef.to_netcdf(filepath)
        except AttributeError:
            # Read as pickle otherwise
            filepath = '{}.pickle'.format(self.get_fit_path(**kwargs))
            if not self.cfg.get('no_verbose'):
                self.log.info('Writing {} {} {} estimator to {}'.format(
                    self.component_mng.name, self.result_mng.name, self.name,
                    filepath))
            # Otherwise, write as pickle
            with open(filepath, 'wb') as f:
                pickle.dump(self.coef, f)


class ExtractorBase(ActuatorBase, ABC):
    """Extractor base class.
    By default, the :py:meth:`transform` method does nothing
    and input data is just read.
    """

    def __init__(self, result_mng, name, cfg=None, variable_names=None,
                 stages={'fit', 'predict', 'output'}, **kwargs):
        """Extractor constructor.

        :param result_mng: Result manager of variable to estimate.
        :param name: Actuator name.
        :param cfg: Estimator configuration. Default is `None`.
        :param variable_names: Name(s) of variable(s) to be extracted.
          Default is `None`.
        :param stages: Stages at which extraction is performed.
          Default is `{'fit', 'predict', 'output'}`.
        :type result_mng: :py:class:`ResultManager`
        :type name: str
        :type cfg: mapping
        :type variable_names: (collection of) :py:class:`str`
        :type stages: (collection of) :py:class:`str`
        """
        #: Estimation stages.
        self.stages = set(stages)

        #: Transformation flag. If `True`, no extraction is performed.
        self.no_extraction = False

        # Initialize as container
        task_names = set()
        for stage in self.stages:
            task_names.update({'extract__' + stage, 'write__' + stage})
        kwargs.update({
            'component_mng': result_mng.component_mng, 'name': name,
            'result_mng': result_mng,
            'cfg': cfg, 'med': result_mng.med, 'task_names': task_names})
        super(ExtractorBase, self).__init__(**kwargs)

        #: Extractor variables.
        self.variable_names = ensure_collection(variable_names, set) or set()

    @abstractmethod
    def transform(self, data_src, stage=None, **kwargs):
        """Abstract transform method."""
        raise NotImplementedError

    def get_extractor_postfix(self, **kwargs):
        """Default implementation: get an empty extractor postfix string.

        :returns: Postfix.
        :rtype: str
        """
        return ''


class DefaultExtractor(ExtractorBase):
    """Default extractor implementation.
    By default, the :py:meth:`transform` method does nothing
    and input data is just read.
    """

    def __init__(self, result_mng, name, cfg=None, **kwargs):
        """Default extractor constructor.

        :param result_mng: Result manager of variable to estimate.
        :param name: Actuator name.
        :param cfg: Estimator configuration. Default is `None`.
        :type result_mng: :py:class:`ResultManager`
        :type name: str
        :type cfg: mapping
        """
        super(DefaultExtractor, self).__init__(
            result_mng=result_mng, name=name, cfg=cfg,
            variable_names=result_mng.result_name, **kwargs)

        # Flag that no transformation is to be performed, unless explicitly
        # asked with `'write'` entry of configuration (for instance if
        # modifier active)
        self.no_extraction = not self.cfg.get('write')

    def transform(self, ds=None, data_src=None, stage=None, **kwargs):
        """Default transform: return equal or modified data source
        for the component and all variables and prevent writing.

        :param data_src: Input data source.
        :param stage: Modeling stage: `'fit'`, `'predict'`, or `'output'`.
        :type data_src: :py:class:`..data_source.DataSourceBase`
        :type stage: str

        :returns: Dataset.
        :rtype: :py:class:`xarray.Dataset`

        .. warning:: If a modifier is active, its 'transform' method
          replaces any such method present in :py:obj:`kwargs['transform']`.
        """
        if hasattr(self.result_mng, 'modifier'):
            if (hasattr(self.result_mng.modifier.cfg, 'stage') and
                    (self.result_mng.modifier.cfg['stage'] != stage)):
                pass
            else:
                # Add modifier transformation to keyword arguments
                kwargs['transform'] = self.result_mng.modifier.transform

        # Get data
        data = data_src.get_data(**kwargs)

        ds = {}
        for variable_name, da in data.items():
            # Try to select component
            try:
                da = da.sel(component=self.component_mng.component_name)
            except ValueError:
                pass

            # Add data array to dataset
            ds[variable_name] = da

        # Prevent writing if same data
        if self.no_extraction:
            self.task_mng['write__' + stage] = False

        # Return equal data for component
        return ds


def add_actuator(container, actuator_name=None, actuator=None,
                 cfg=None, actuator_class_name='Actuator', **kwargs):
    """Add actuator (extractor, estimator, modifier, etc.).

    :param container: Container to which to add actuator.
    :param actuator_name: Actuator name. Default is `None`,
      in which case a `TypeError` will be raised if `actuator` is
      not `None`.
    :param actuator: Actuator. Default is `None`, in which case
      it is built with :py:meth:`_init_actuator`.
    :param cfg: Configuration in which to look for actuator configurations.
      Default is `None`, in which case it is taken from the parent
      configuration.
    :param actuator_class_name: Actuator class to call from module.
      Default is `'Actuator'`.
    :type container: :py:class:`.container.Container`
    :type actuator_name: str
    :type actuator: :py:class:`.container.Container`
    :type cfg: mapping
    :type actuator_class_name: str
    """
    cfg = cfg or container.cfg
    if actuator is None:
        if actuator_name is None:
            msg = "missing 1 required positional argument: 'actuator_name'"
            raise TypeError(msg)
        if actuator_name in cfg:
            actuator_cfg = cfg[actuator_name]
            actuator_module = import_module_from_cfg(
                actuator_cfg, name=actuator_name)

            if actuator_module is None:
                # Get default actuator
                actuator_class_name = 'Default' + ''.join(
                    act_name.title()
                    for act_name in actuator_name.split('_'))
                actuator_module = sys.modules[__name__]
                container.log.warning(
                    '    No "module_path" found in {} configuration: '
                    'using {} class actuator'.format(
                        actuator_name, actuator_class_name))

            # Try to get actuator
            if hasattr(actuator_module, actuator_class_name):
                actuator = getattr(actuator_module, actuator_class_name)(
                    container, actuator_name, cfg=actuator_cfg)
            else:
                raise AttributeError(
                    'No {} class found in {} module for {} {} '
                    'actuator'.format(
                        actuator_class_name, actuator_module,
                        container.name, actuator_name))

    if actuator is not None:
        if not container.cfg.get('no_verbose'):
            container.log.info('Injecting {} actuator in {} {}'.format(
                actuator.name, container.component_mng.name, container.name))
        return actuator


# Alias default feature and output extractor
DefaultFeatureExtractor = DefaultOutputExtractor = DefaultExtractor



"""Bias corrector and predictor."""
from numpy import set_printoptions
import numpy as np
import xarray as xr
from sklearn.exceptions import NotFittedError
from sklearn import linear_model
from ..actuator_base import EstimatorBase
from ..utils.learn import parse_fit_data, select_data

REGRESSOR = linear_model.RidgeCV


class Actuator(EstimatorBase):
    """Bias corrector and predictor."""

    def __init__(self, result_mng, name, cfg=None, **kwargs):
        """Naming constructor.

        :param result_mng: Result manager of variable to estimate.
        :param name: Actuator name.
        :param cfg: Actuator configuration. Default is `None`.
        :type result_mng: :py:class:`ResultManager`
        :type name: str
        :type cfg: mapping
        """
        super(Actuator, self).__init__(
            result_mng=result_mng, name=name, cfg=cfg,
            **kwargs)

    def fit(self, data_src_in, data_src_out, **kwargs):
        """Get bias corrector for component as a factor to multiply the
        input data with and such that the training input data
        multiplied by the bias corrector has the same mean
        as the training output data over the same period.

        :param data_src_in: Training set input data source.
        :param data_src_out: Training set output data source.
        :type data_src_in: :py:class:`..data_source.DataSourceBase`
        :type data_src_out: :py:class:`..data_source.DataSourceBase`
        """
        # Get input and output data arrays
        da_in, da_out = parse_fit_data(
            data_src_in, data_src_out, self.result_mng,
            subsample_freq=self.cfg.get('subsample_freq'),
            select_period=True, time_slice=self.cfg.get('time_slice'),
            **kwargs)

        # Select intersection
        t_in = da_in.indexes['time']
        t_intersect = t_in.intersection(da_out.indexes['time'])
        da_in = da_in.sel(time=t_intersect)
        da_out = da_out.sel(time=t_intersect)

        # Select valid
        valid = ~da_in.isnull() & ~da_out.isnull()
        da_in = da_in.where(valid, drop=True)
        da_out = da_out.where(valid, drop=True)

        if not self.cfg.get('no_verbose'):
            self.log.info(
                'Computing {} {} bias corrector from {} to {} '.format(
                    self.result_mng.name, self.component_mng.name,
                    *da_in.indexes['time'][[0, -1]]))

        # Make multiple regional models into one model
        X = _get_factorized_input(da_in, **kwargs)
        y = _get_factorized_output(da_out, **kwargs)

        # Configure regression
        regressor_kwargs = self.cfg.get('regressor_kwargs') or {}

        if 'cv' not in regressor_kwargs:
            # Split by year for cross validation
            years = da_in.indexes['time'].year
            regressor_kwargs['cv'] = years.max() - years.min() + 1

        # Prevent fitting intercept
        regressor_kwargs['fit_intercept'] = False

        if 'alphas' not in regressor_kwargs:
            # Define regularization path from given logspace parameters
            alphas_logspace = self.cfg.get('alphas_logspace')
            if alphas_logspace is not None:
                alphas = np.logspace(**alphas_logspace)
                regressor_kwargs['alphas'] = alphas

        # Define regressor
        reg = REGRESSOR(**regressor_kwargs)

        # Ridge regression with cross validation
        reg.fit(X, y)

        # Get score
        score = reg.score(X, y)

        # Save coefficients
        try:
            coords = [da_in.region]
        except AttributeError:
            coords = []
        self.coef = xr.DataArray(reg.coef_, coords=coords, name='coef')
        self.coef.attrs['score'] = score
        self.coef.attrs['cv'] = regressor_kwargs['cv']
        try:
            self.coef.attrs['alpha'] = reg.alpha_
            self.coef.attrs['alphas'] = reg.alphas_
        except AttributeError:
            pass

        if not self.cfg.get('no_verbose'):
            set_printoptions(precision=3)
            self.log.info('Scale regression coefficients:')
            self.log.info(self.coef)

    def predict(self, data_src_in, **kwargs):
        """Apply bias corrector for component by multipling input data.

        :param data_src_in: Input data source.
        :type data_src_in: :py:class:`..data_source.DataSourceBase`

        :returns: Bias corrected dataset.
        :rtype: dict
        """
        # Verify that bias corrector has been fitted
        if self.coef is None:
            raise NotFittedError('This bias corrector instance '
                                 'must be fitted before prediction.')

        # Select result, copy and ensure dimensions order
        da_in = select_data(data_src_in, self.result_mng, **kwargs)

        # Reorder regions to comply
        coef_comp = (self.coef.loc[{'region': da_in.indexes['region']}]
                     if 'region' in self.coef.coords else self.coef.copy())

        # Select component if needed
        try:
            coef_comp = coef_comp.sel(
                component=self.component_mng.component_name)
        except ValueError:
            pass

        # Define regressor with previously fitted coefficients
        regressor_kwargs = self.cfg['regressor_kwargs'] or {}

        # Prevent fitting intercept
        regressor_kwargs['fit_intercept'] = False

        reg = REGRESSOR(**regressor_kwargs)
        reg.intercept_ = 0.

        # Set fitted coefficients
        reg.coef_ = self.coef.values

        # Initialize prediction depending on input dimensions
        if len(da_in.shape) > 1:
            da_pred = xr.full_like(
                da_in.stack(stacked_dim=['region', 'time']), None)
        else:
            da_pred = xr.full_like(da_in, None)

        # Get factorized input
        X = _get_factorized_input(da_in, **kwargs)

        # Predict
        da_pred[:] = reg.predict(X)

        # Unstack, if needed
        if len(da_in.shape) > 1:
            da_pred = da_pred.unstack('stacked_dim').transpose(
                'time', 'region')

        return {self.result_mng.result_name: da_pred}

    def get_estimator_postfix(self, **kwargs):
        """Get bias-corrector postfix.

        returns: Postfix.
        rtype: str
        """
        return '{}_scaled'.format(
            super(Actuator, self).get_estimator_postfix(**kwargs))


def _get_factorized_input(da_in, **kwargs):
    """Get input matrix factorized over regions.

    :param da_in: Input matrix.
    :type da_in: :py:class:`xarray.DataArray`

    :returns: Factorized input matrix.
    :rtype: :py:class:`numpy.ndarray`
    """
    shape = da_in.shape
    if len(shape) == 1:
        X = da_in.values
    else:
        nt, nreg = shape
        X = np.zeros((nt * nreg, nreg))
        for ir in range(nreg):
            # Set block of rows and column corresponding to region
            sl = slice(ir * nt, (ir + 1) * nt)

            # Linear
            X[sl, ir] = da_in.isel(region=ir).values

    return X


def _get_factorized_output(da_out, **kwargs):
    """Get output vector factorized over regions.

    :param da_out: Output matrix.
    :type da_out: :py:class:`xarray.DataArray`

    :returns: Factorized output vector.
    :rtype: :py:class:`numpy.ndarray`
    """
    if len(da_out.shape) == 1:
        # Return values
        y = da_out.values
    else:
        # Stack blocks of rows associated to each region
        y = da_out.stack(stacked_dim=['region', 'time'])

    return y



"""Eirgrid API for Ireland."""
import os
import requests
import numpy as np
import pandas as pd
import xarray as xr
from ..data_source import DataSourceLoaderBase
from ..component import (
    DataSourceWithComponentsMixin, parse_variable_component_args,
    download_to_compute_capacity_factor, compute_capacity_factor,
    finalize_dataset)


class DataSource(DataSourceLoaderBase, DataSourceWithComponentsMixin):
    """Eirgrid data source."""
    #: Default source name.
    DEFAULT_SRC_NAME = 'eirgrid'

    #: Area: `'Ireland'`
    AREA = 'Ireland'

    def __init__(self, med, name=None, cfg=None, **kwargs):
        """Naming constructor.

        :param med: Mediator.
        :param name: Data source name.
        :param cfg: Data source configuration. Default is `None`.
        :type med: :py:class:`.mediator.Mediator`
        :type name: str
        :type cfg: mapping
        """
        name = name or self.DEFAULT_SRC_NAME
        super(DataSource, self).__init__(med, name, cfg=cfg, **kwargs)

    @parse_variable_component_args
    @download_to_compute_capacity_factor
    def download(self, variable_component_names, **kwargs):
        """Download data.

        :param variable_component_names: Names of components to download per
          variable.
        :type variable_component_names: mapping from :py:class:`str`
          to :py:class:`set` of :py:class:`str`

        :returns: Names of downloaded components per variable.
        :rtype: mapping from :py:class:`str` to :py:class:`set` of
          :py:class:`str`

        .. note:: The data is downloaded from
          `Eirgrid <http://www.eirgridgroup.com/how-the-grid-works/renewables/>`.
        """
        # Loop over variables
        for variable_name, component_names in variable_component_names.items():
            if not self.cfg.get('no_verbose'):
                self.log.info('- {}:'.format(variable_name))

            # Manage wind component as onshore only
            component_name = 'wind'
            component_to_download_names = component_names.copy()
            if component_name in component_names:
                # Make sure that generation and capacity are loaded
                component_to_download_names.add('wind-onshore')

                # Remove capacity factor from components to load
                component_to_download_names.discard(component_name)

            # Loop over year groups
            dates = self._get_dates(variable_name, **kwargs)
            for date in dates:
                if (not self.cfg.get('no_verbose')) and (date is not None):
                    self.log.info('-- {}'.format(date))

                # Loop over components
                for component_name in component_to_download_names:
                    # Get URL
                    url, filepath = self._get_url_filepath(
                        variable_name, component_name, date, **kwargs)
                    if not self.cfg.get('no_verbose'):
                        self.log.info('--- {} from {} to {}'.format(
                            component_name, url, filepath))

                    # Request and raise exception if needed
                    response = requests.get(url)
                    response.raise_for_status()

                    # Write XLSX
                    with open(filepath, 'wb') as f:
                        f.write(response.content)

        # Return names of downloaded variables
        return variable_component_names

    @parse_variable_component_args
    @compute_capacity_factor
    @finalize_dataset
    def load(self, variable_component_names, **kwargs):
        """Load data.

        :param variable_component_names: Names of components to load per
          variable
        :type variable_component_names: mapping from :py:class:`str`
          to collection

        :returns: Time series for each variable and component.
        :rtype: dict
        """
        # Get place names for source
        src_place_regions, _ = (
            self.med.geo_src.get_place_regions_for_source(
                self.name, self.AREA, **kwargs))
        place_names = list(src_place_regions)

        # Loop over variables
        ds = {}
        start = pd.Timestamp('{}-01-01'.format(self.cfg['first_year']))
        end = pd.Timestamp('{}-01-01'.format(self.cfg['last_year'] + 1))
        time_slice = slice(start, end)
        for variable_name, component_names in variable_component_names.items():
            cfg_var = self.cfg[variable_name]
            if not self.cfg.get('no_verbose'):
                self.log.info('- {}:'.format(variable_name))

            # Manage wind component as onshore only
            component_name = 'wind'
            component_to_load_names = component_names.copy()
            if component_name in component_names:
                # Make sure that generation and capacity are loaded
                component_to_load_names.add('wind-onshore')

                # Remove capacity factor from components to load
                component_to_load_names.discard(component_name)

            # Loop over year groups
            dates = self._get_dates(variable_name, **kwargs)
            da = None
            for date in dates:
                if (not self.cfg.get('no_verbose')) and (date is not None):
                    self.log.info('-- {}'.format(date.year))

                # Loop over components
                da_date = None
                for component_name in component_to_load_names:
                    # Get filepath
                    _, filepath = self._get_url_filepath(
                        variable_name, component_name, date, **kwargs)
                    if not self.cfg.get('no_verbose'):
                        self.log.info('-- {} from {}'.format(
                            component_name, filepath))

                    # Read data frame
                    df_comp = pd.read_excel(
                        filepath, **cfg_var['read_excel_kwargs'])

                    if variable_name == 'generation':
                        # Convert time zone to UTC
                        df_comp.index = df_comp.index.tz_localize(
                            self.cfg['timezone'], ambiguous=df_comp[
                                'GMT Offset'].astype(bool)).tz_convert(None)
                    df_comp.index.name = 'time'

                    # Select variable
                    df_comp = df_comp[cfg_var['variable_names']
                                      [component_name]]

                    # Convert to data array
                    da_comp = xr.DataArray(
                        df_comp, dims=('time',)).expand_dims(
                        'component').assign_coords(component=[component_name])

                    # Add component to data array
                    da_date = (da_comp if da_date is None else
                               xr.concat([da_date, da_comp], dim='component'))

                # Add component to data array
                da = (da_date if da is None else
                      xr.concat([da, da_date], dim='time'))

            # Select period
            da = da.sel(time=time_slice)

            if variable_name == 'generation':
                # Convert 15m-power to 1h-energy
                da = da.resample(time='H').mean('time')

            # Add area as region coordinate
            da = da.expand_dims('region').assign_coords(
                region=place_names)

            # Add total wind component as onshore wind if needed
            component_name = 'wind'
            if component_name in component_names:
                coord_component = da.indexes['component'].tolist()
                idx = coord_component.index('wind-onshore')
                coord_component[idx] = component_name
                da['component'] = coord_component

            # Add to dataset
            ds[variable_name] = da

        return ds

    def _get_capacity_factor(self, ds, **kwargs):
        """Get capacity factor from generation and capacity.

        :param ds: Dataset containing generation and capacity.
        :type ds: mapping

        :returns: Capacity factor.
        :rtype: :py:class:`xarray.DataArray`
        """
        da_gen = ds['generation']
        da_cap = ds['capacity']

        # Upsample capacity to generation
        freq = da_gen.indexes['time'].inferred_freq
        da_cap_up = da_cap.resample(time=freq).ffill().reindex(
            time=da_gen.indexes['time'], method='ffill')

        # Get number of hours in generation time step
        nhours = 1.

        # Get capacity factor
        da = da_gen / (da_cap_up * nhours)

        return da

    def _get_dates(self, variable_name, **kwargs):
        """ Get dates from which to get year groups over which to loop.

        :param variable_name: variable_name.
        :type variable_name: str

        :returns: Dates.
        :rtype: :py:class:`pandas.DatetimeIndex`
        """
        if 'year_groups' in self.cfg[variable_name]:
            start_year = self._get_year_group(
                variable_name, self.cfg['first_year'])[0]
            start = pd.Timestamp('{}-12-31'.format(start_year))
            end_year = self._get_year_group(
                variable_name, self.cfg['last_year'])[-1]
            end = pd.Timestamp('{}-12-31'.format(end_year))
            dates = pd.date_range(
                start=start, end=end, freq='2A-DEC')
        else:
            dates = [None]

        return dates

    def _get_year_group(self, variable_name, year, **kwargs):
        """Get year group to which :py:obj:`year` belong.

        :param variable_name: Variable name.
        :param year: Year.
        :type variable_name: str
        :type year: int

        :returns: Year group.
        :rtype: list
        """
        year_groups = self.cfg[variable_name]['year_groups']
        year_group = np.array(year_groups)[
            [year in yg for yg in year_groups]][0]

        return year_group

    def _get_url_filepath(self, variable_name, component_name, date=None,
                          **kwargs):
        """Get URL for query and filepath.

        :param variable_name: Name of variable to query.
        :param component_name: Name of component to query.
        :param date: Date from which to get year.
          Required for 'generation' variable. Default is `None`.
        :type variable_name: str
        :type component_name: str
        :type date: :py:class:`pandas.Timestamp`

        :returns: Tuple containing URL and filepath.
        :rtype: :py:class:`tuple` with two :py:class:`str`
        """
        # Define URL and filename
        if component_name in ['wind', 'wind-onshore', 'wind-offshore']:
            cfg_var = self.cfg[variable_name]
            if variable_name == 'capacity':
                # Define filename for capacity
                filename = '{}.{}'.format(
                    cfg_var['filename'], cfg_var['file_format'])
            elif variable_name == 'generation':
                # Define filename for generation
                year_group = self._get_year_group(
                    variable_name, date.year, **kwargs)
                filename = '{}-{}-{}.{}'.format(
                    cfg_var['filename'], *year_group, cfg_var['file_format'])

            # URL
            url = '{}/{}'.format(self.cfg['host'], filename)

        # Get filepath
        src_dir = self.med.cfg.get_external_data_directory(self)
        filepath = os.path.join(src_dir, filename)

        return url, filepath

    def get_postfix(self, **kwargs):
        """Get data-source standard postfix.

        :returns: Postfix.
        :rtype: str
        """
        postfix = '_{}-{}'.format(self.cfg['first_year'],
                                  self.cfg['last_year'])

        return postfix



"""Italian calendar."""
from collections import OrderedDict
import numpy as np
import xarray as xr
from ..data_source import DataSourceLoaderBase

#: Default data-source name;
DEFAULT_SRC_NAME = 'calendar'

#: Default variable name.
VARIABLE_NAME = 'calendar'


class DataSource(DataSourceLoaderBase):
    """Calendar for Italy."""

    def __init__(self, med, name=None, cfg=None, **kwargs):
        """Initialize Italian calendar as data_source.

        :param med: Mediator.
        :param name: Data source name.
        :param cfg: Data-source configuration. Default is `None`.
        :type med: :py:class:`.mediator.Mediator`
        :type name: str
        :type cfg: mapping
        """
        name = name or DEFAULT_SRC_NAME
        super(DataSource, self).__init__(med, name, cfg=cfg, **kwargs)

    def download(self, *args, **kwargs):
        """Convenience function to warn that not calendar data needs to be
        downloaded.

        :returns: Names of downloaded variable.
        :rtype: :py:class:`set` of :py:class:`str`
        """
        self.log.warning(
            '{} data need not be downloaded'.format(self.name))

        return {VARIABLE_NAME}

    def load(self, dl, **kwargs):
        """Get array of working days, holidays and saturdays for Italy,
        and store it to member :py:attr::`data`.

        :param dl: List of dates
        :type dl: pd.DatetimeIndex, datetime.datetime

        :returns: Calendar dataset.
        :rtype: :py:class:`xarray.Dataset`
        """
        # Create day type coordinate
        daytype_name_idx = OrderedDict({'work': 0, 'sat': 1, 'off': 2})
        daytype_coord = ('daytype_index', list(daytype_name_idx.values()))

        # Create array
        da = xr.DataArray(np.empty((dl.shape[0],), dtype=int),
                          coords=[('time', dl)])

        # Sundays
        off = (dl.dayofweek == 6)
        # Public holidays + Winter and Summer holidays
        # 1st of January: new year day
        a = (dl.dayofyear == 1)
        # 25th of April: liberation day
        b = ((dl.month == 4) & (dl.day == 25))
        # 1st of May: May Day
        c = ((dl.month == 5) & (dl.day == 1))
        # 8th to 21st of August: summer holidays
        d = ((dl.month == 8) & (dl.day > 6)
             & (dl.day < 22))
        # 1st of November: all saints
        e = ((dl.month == 11) & (dl.day == 1))
        # 8th of December: Feast of the Immaculate Conception
        f = ((dl.month == 12) & (dl.day == 8))
        # 25th to 31st of December: Christmas holidays
        g = ((dl.month == 12) & (dl.day >= 25))

        # Concatenate off days
        off = (off | a | b | c | d | e | f | g)
        da[off] = daytype_name_idx['off']

        # Saturdays
        sat = (dl.dayofweek == 5) & ~off
        da[sat] = daytype_name_idx['sat']

        # Working Days
        da[~sat & ~off] = daytype_name_idx['work']

        # Create dataset with daytype coordinate
        ds = xr.Dataset({VARIABLE_NAME: da})
        ds = ds.assign_coords(daytype=xr.DataArray(
            list(daytype_name_idx.keys()), coords=[daytype_coord]))

        return ds



"""Extractor composing multiple extractors."""
from collections import OrderedDict
from ..actuator_base import ExtractorBase, add_actuator
from ..data_source import Composer


class Actuator(ExtractorBase):

    def __init__(self, result_mng, name, cfg=None, **kwargs):
        """Naming constructor.

        :param result_mng: Result manager of variable to estimate.
        :param name: Actuator name.
        :param cfg: Actuator configuration. Default is `None`.
        :type result_mng: :py:class:`ResultManager`
        :type name: str
        :type cfg: mapping

        .. note:: Actuators to compose must be provided as a dictionary
          from the `extractors` entry of the configuration.
        """
        super(Actuator, self).__init__(
            result_mng=result_mng, name=name, cfg=cfg, **kwargs)

        # Initialize actuators from configuration
        self.actuators = OrderedDict()
        for actuator_name in self.cfg['extractors']:
            # Get actuator
            actuator = add_actuator(
                self.result_mng, actuator_name=actuator_name,
                cfg=self.cfg['extractors'], **kwargs)

            # Reference actuator in actuators set
            self.actuators[actuator_name] = actuator

        #: Functions composed to define :py:meth:`transform` method.
        self._functions = None

        # Add transform of each actuators in the right order
        self._functions = [self.actuators[actuator_name].transform
                           for actuator_name in self.cfg['extractors']]

        # Define :py:meth:`transform` method by composing functions
        self.transform = Composer(*self._functions)

        # Add data sources
        for actuator in self.actuators.values():
            if 'data' in actuator.cfg:
                if self.data_sources is None:
                    self.data_sources = OrderedDict()
                self.add_data_sources(cfg=actuator.cfg, **kwargs)

    def transform(self, **kwargs):
        """Empty concrete definition of transform method replaced
        by composed functions during construction."""
        pass

    def get_extractor_postfix(self, **kwargs):
        """Get extractor postfix.

        returns: Postfix.
        rtype: str
        """
        # Get default postfix
        postfix0 = super(Actuator, self).get_extractor_postfix(**kwargs)

        # Add postfixes of every actuator
        return ''.join([postfix0] + [
            self.actuators[actuator_name].get_extractor_postfix(**kwargs)
            for actuator_name in self.cfg['extractors']])



"""RTE ECO2MIX API."""
import os
import requests
import zipfile
from io import BytesIO
import pandas as pd
import xarray as xr
from ..data_source import DataSourceLoaderBase
from ..component import (DataSourceWithComponentsMixin,
                         parse_variable_component_args)


class DataSource(DataSourceLoaderBase, DataSourceWithComponentsMixin):
    #: Default source name.
    DEFAULT_SRC_NAME = 'eco2mix'

    #: Area: `'France'`.
    AREA = 'France'

    def __init__(self, med, name=None, cfg=None, **kwargs):
        """Naming constructor.

        :param med: Mediator.
        :param name: Data source name.
        :param cfg: Data source configuration. Default is `None`.
        :type med: :py:class:`.mediator.Mediator`
        :type name: str
        :type cfg: mapping
        """
        name = name or self.DEFAULT_SRC_NAME
        super(DataSource, self).__init__(med, name, cfg=cfg, **kwargs)

    def download(self, **kwargs):
        """Download data.

        :returns: Names of downloaded components per variable.
        :rtype: mapping from :py:class:`str` to :py:class:`set` of
          :py:class:`str`

        .. note:: The data is downloaded from the
          `eco2mix website <https://eco2mix.rte-france.com/download/eco2mix/>`.
        """
        # Get region names
        _, src_region_place = (
            self.med.geo_src.get_place_regions_for_source(
                self.name, self.AREA, **kwargs))

        # Get data-source directory
        src_dir = self.med.cfg.get_external_data_directory(self)

        years = range(int(self.cfg['first_year']),
                      int(self.cfg['last_year']) + 1)
        for year in years:
            if not self.cfg.get('no_verbose'):
                self.log.info('Downloading files for {}'.format(year))
            for region_name in src_region_place:
                url = self._get_url(region_name, year, **kwargs)
                if not self.cfg.get('no_verbose'):
                    self.log.info('  {}'.format(url))

                # Request and raise exception if needed
                response = requests.get(url)
                response.raise_for_status()

                # Unzip
                zip_ref = zipfile.ZipFile(BytesIO(response.content))
                zip_ref.extractall(src_dir)
                zip_ref.close()

        return {'generation': {'wind-onshore', 'pv'},
                'capacity': {'wind-onshore', 'pv'},
                'capacity_factor': {'wind-onshore', 'pv'}}

    @parse_variable_component_args
    def load(self, variable_component_names, **kwargs):
        """Load data.

        :param variable_component_names: Names of components to load per
          variable.
        :type variable_component_names: mapping from :py:class:`str`
          to collection

        :returns: Time series for each variable and component.
        :rtype: :py:class:`xarray.Dataset()`
        """
        # Manage wind component as onshore only
        component_name = 'wind'
        component_names = variable_component_names[
            list(variable_component_names)[0]]
        component_to_load_names = component_names.copy()
        if component_name in component_names:
            # Make sure that generation and capacity are loaded
            component_to_load_names.add('wind-onshore')

            # Remove capacity factor from components to load
            component_to_load_names.discard(component_name)

        # Get region-place mapping
        _, src_region_place = self.med.geo_src.get_place_regions_for_source(
            self.name, self.AREA, **kwargs)

        # Get data-source directory
        src_dir = self.med.cfg.get_external_data_directory(self)

        # Get years
        years = range(int(self.cfg['first_year']),
                      int(self.cfg['last_year']) + 1)
        da = None
        for year in years:
            if not self.cfg.get('no_verbose'):
                self.log.info('Reading files for {}'.format(year))
            start_date = '{}-01-01 00:00'.format(year)
            end_date = '{}-12-31 23:45'.format(year)
            da_year = None
            for region_name, place_name in src_region_place.items():
                # Read CSV file
                filename = self._get_download_filename(
                    year, region_name, self.cfg['source_file_format'])
                filepath = os.path.join(src_dir, filename)
                if not self.cfg.get('no_verbose'):
                    self.log.info('  {}'.format(filepath))
                df = pd.read_csv(filepath, **self.cfg['read_csv_kwargs'])

                # Add datetime index
                df.index = pd.date_range(
                    start=start_date, end=end_date,
                    freq=self.cfg['frequency'])
                # df.index = pd.to_datetime(list(
                #     '{}\t{}'.format(d, h)
                #     for d, h in df[['Date', 'Heures']].values))
                df.index.name = 'time'

                # Drop columns
                for col in self.cfg['drop_columns']:
                    del df[col]

                # Rename components and select them
                component_names_inv = {
                    self.cfg['component_names'][component_name]:
                    component_name
                    for component_name in component_to_load_names}
                df = df.rename(columns=component_names_inv)[
                    component_to_load_names]

                # Convert 'ND' and '-' values to None
                is_null = (df == 'ND') | (df == '-')
                if is_null.any().any():
                    df[is_null] = None

                # Convert to float
                df = df.astype(float)

                # Get array with places as coordinates to groupby
                da_reg = xr.DataArray(df, dims=['time', 'component'])
                da_reg = da_reg.expand_dims(
                    'region').assign_coords(region=[place_name])

                # Concatenate regions
                da_year = da_reg if da_year is None else xr.concat(
                    [da_year, da_reg], dim='region')

            # Group by places summing energies
            da_year = da_year.groupby('region').sum('region')

            # Concatenate years
            da = da_year if da is None else xr.concat(
                [da, da_year], dim='time')

        # Sub-sample to hourly
        da = da.resample(time='H').sum('time')

        # Add total wind component as onshore wind if needed
        component_name = 'wind'
        if component_name in component_names:
            coord_component = da.indexes['component'].tolist()
            idx = coord_component.index('wind-onshore')
            coord_component[idx] = component_name
            da['component'] = coord_component

        # Create dataset of variables
        ds = xr.Dataset()
        variable_name = 'demand'
        if variable_name in variable_component_names:
            # Add demand variables with demand component only
            ds[variable_name] = self.finalize_array(
                da.sel(component='demand'), variable_name, **kwargs)
        variable_name = 'generation'
        if variable_name in variable_component_names:
            # Add generation variable with generation components only
            gen_comp = list(da.indexes['component'])
            try:
                gen_comp.pop(gen_comp.index('demand'))
            except ValueError:
                pass
            ds[variable_name] = self.finalize_array(
                da.sel(component=gen_comp), variable_name, **kwargs)

        return ds

    def get_postfix(self, **kwargs):
        """Get data-source standard postfix.

        :returns: Postfix.
        :rtype: str
        """
        postfix = '_{}-{}'.format(
            self.cfg['first_year'], self.cfg['last_year'])

        return postfix

    def _get_url(self, region_name, year, **kwargs):
        """Get URL for region and year.

        :param region_name: Region name.
        :param year: Year.
        :type region_name: str
        :type year: int

        :returns: URL.
        :rtype: str
        """
        filename = self._get_download_filename(year, region_name, 'zip')
        url = os.path.join(self.cfg['host'], filename)

        return url

    def _get_download_filename(self, year, region_name, file_format, **kwargs):
        """Get name of file to download/downloaded.

        :param year: Year.
        :param region_name: French administrative region.
        :param file_format: File format.
        :type year: str
        :type region_name: str
        :type file_format: str

        :returns: Filename.
        :rtype: str
        """
        region_file = self.cfg['region_files'][region_name]
        filename = 'eCO2mix_RTE_{}_Annuel-Definitif_{}.{}'.format(
            region_file, year, file_format)

        return filename



"""Optimization base definitions."""
from abc import ABC, abstractmethod
from .container import Container, load_config
from .data_source import DataSourceLoaderBase


class OptimizerBase(Container, ABC):
    """Optimizer abstract base class. Requires :py:meth:`solve` method to be
    implemented."""

    def __init__(self, med, name, cfg=None, **kwargs):
        """Build optimizer linked to mediator.

        :param med: Mediator.
        :param name: Optimizer name.
        :param cfg: Optimizer configuration.
          If `None`, call :py:meth:`.config._load_src_coonfig`.
          Default is `None`.
        :type med: :py:class:`.mediator.Mediator`
        :type name: str
        :type cfg: mapping
        """
        # Load optimizer configuration
        loaded_cfg = cfg or load_config(self.med, self.name)

        # Initialize as container with mediator as parent
        super(OptimizerBase, self).__init__(
            med, name, cfg=loaded_cfg, parent=med, **kwargs)

        #: Input data source.
        self.input = None

        #: Solution data source.
        self.solution = SolutionBase(self, **kwargs)

    @abstractmethod
    def solve(self, *args, **kwargs):
        """Abstract method to solve optimization problem."""
        raise NotImplementedError


class InputBase(DataSourceLoaderBase):
    """Abstract optimization input class as data source with loader.
    Requires :py:meth:`load` method inherited from
    :py:class:`.data_source.DataSourceLoaderBase` to be implemented."""

    def __init__(self, optimizer, cfg=None, **kwargs):
        """Initialize input data source.

        :param optimizer: Optimizer.
        :param cfg: Input configuration. Default is `None`.
        :type optimizer: :py:class:`.optimization.OptimizerBase`
        :type cfg: mapping
        """
        self.optimizer = optimizer
        name = '{}_input'.format(self.optimizer.name)

        super(InputBase, self).__init__(
            self.optimizer.med, name, cfg=cfg, parent=self.optimizer, **kwargs)

    def get_data_postfix(self, **kwargs):
        """Get data postfix.

        :returns: Postfix.
        :rtype: str
        """
        return ''

    def get_data_dir(self, makedirs=True, **kwargs):
        """Get path to data directory.

        :param makedirs: Make directories if needed. Default is `True`.
        :type makedirs: bool

        :returns: Data directory path.
        :rtype: str
        """
        return self.med.cfg.get_project_data_directory(
            self.optimizer, subdirs='input', makedirs=makedirs)


class SolutionBase(DataSourceLoaderBase):
    """Optimization solution base class as data source."""

    def __init__(self, optimizer, cfg=None, **kwargs):
        """Initialize solution as data source.

        :param optimizer: Optimizer.
        :param cfg: Solution configuration. Default is `None`.
        :type optimizer: :py:class:`.optimization.OptimizerBase`
        :type cfg: mapping
        """
        self.optimizer = optimizer
        name = '{}_solution'.format(self.optimizer.name)

        super(SolutionBase, self).__init__(
            self.optimizer.med, name, cfg=cfg, parent=self.optimizer, **kwargs)

    def load(self, variable_names=None, **kwargs):
        """Load returning result from :py:meth:`OptimizerBase.solve`.

        :param variable_names: Names of variables to download.
          Default is `None`, in which case all variables in
          :py:attr`variables` are downloaded.
        :type variable_names: (collection of) :py:class:`str`
        """
        # Get input data
        self.optimizer.input.get_data(**kwargs)

        # Solve
        return self.optimizer.solve(variable_names=variable_names, **kwargs)

    def get_data_dir(self, makedirs=True, **kwargs):
        """Get path to data directory.

        :param makedirs: Make directories if needed. Default is `True`.
        :type makedirs: bool

        :returns: Data directory path.
        :rtype: str
        """
        return self.med.cfg.get_project_data_directory(
            self.optimizer, subdirs='solution', makedirs=makedirs)

    def get_data_postfix(self, **kwargs):
        """Default implementation of get optimization results postfix.

        :returns: Postfix.
        :rtype: str
        """
        return self.optimizer.input.get_data_postfix()



"""Mediator configuration management."""
import os
from collections import MutableMapping
from importlib import import_module
import oyaml as yaml
from getpass import getpass
from .container import ensure_collection


def input_expand_user(fun):
    """Decorator of :py:func:`input` to manage `'~'` in filepath."""
    def decorated_fun(prompt):
        q = fun(prompt)
        return os.path.expanduser(q)

    return decorated_fun


# Decorate open to manage ~ in paths with decorator_expand_user
input = input_expand_user(input)


class Config(MutableMapping):
    """Configuration class."""

    def __init__(self, pre_config_filepath=None):
        """Load configuration file.

        :param pre_config_filepath: Filepath of (pre-)configuration file.
          If `None`, ask user. Default is `None`.
        :type pre_config_filepath: str
        """
        #: Actual configuration.
        self.cfg = None

        if not pre_config_filepath:
            # Ask for the configuation file
            default_cfg_filepath = os.path.join('cfg', 'which_config.yaml')
            pre_config_filepath = input(
                'Mediator configuration filepath ([{}]): '.format(
                    default_cfg_filepath))
            if not pre_config_filepath:
                pre_config_filepath = default_cfg_filepath

        # Open pre-configuration file
        with open(pre_config_filepath, 'r') as f:
            precfg = yaml.load(f, Loader=yaml.FullLoader)

        if 'cfg_filepath' in precfg:
            # If pre-configuration file has a configuration filename,
            # load the target configuration file
            cfg_filepath = precfg['cfg_filepath']
            print('Loading configuration from '.format(cfg_filepath))
            with open(cfg_filepath, 'r') as f:
                self.cfg = yaml.load(f, Loader=yaml.FullLoader)
        else:
            # Else, return the pre-configuration
            self.cfg = precfg

    def get_credentials(self, name, keys=['user', 'passwd']):
        """Get credentials given by :py:obj:`keys` argument
        for a container named :py:obj:`name` from file, or ask them.

        :param name: Container name.
        :param keys: Name of asked credentials.
        :type name: str
        :type keys: list

        :returns: Credential key-value pairs.
        :rtype: dict
        """
        # Get credentials
        keys_path = os.path.join(*ensure_collection(
            self.cfg['cfg_path']['keys']))
        if os.path.isfile(keys_path):
            # Try to get them from keys file
            with open(keys_path, 'r') as f:
                cfg_keys = yaml.load(f, Loader=yaml.FullLoader)

            # Get or create container credentials dictionary
            if not isinstance(cfg_keys.get(name), MutableMapping):
                cfg_keys[name] = {}
            cred = cfg_keys[name]
        else:
            cfg_keys = {name: {}}
            cred = {}

        any_new = False
        for k in keys:
            if cred.get(k) is None:
                # Get credential value for key
                cred[k] = getpass('{} {}: '.format(name, k))
                any_new = True

        if any_new:
            # Ask if user wants to save keys to keys file
            q = None
            while q not in ['no', 'yes', '']:
                q = input(
                    'Save new redentials to {} ([yes]/no)? '.format(
                        keys_path))
            if q in ['yes', '']:
                with open(keys_path, 'r+') as f:
                    yaml.dump(cfg_keys, f, default_flow_style=False)

        # Return only asked key-value pairs
        return {k: cred.get(k) for k in keys}

    def get_data_root_directory(self, **kwargs):
        """Get data root directory from mediator configuration.

        :returns: Data root-directory.
        :rtype: str
        """
        directory = self.get('data_dir')
        directory = ('data' if directory is None else
                     os.path.join(*ensure_collection(directory)))

        return directory

    def get_project_data_directory(
            self, container=None, makedirs=True, subdirs=None, **kwargs):
        """Get project data directory and create it if needed.

        :param container: Container for which to get data directory.
        :param makedirs: Make directories if needed. Default is `True`.
        :param subdirs: List of additional subdirectories.
        :type container: :py:class:`.container.Container`
        :type makedirs:
        :type subdirs: :py:class:`str` or :py:class:`list` or :py:class:`str`

        :returns: Directory path.
        :rtype: str
        """
        subdirs = ensure_collection(subdirs)
        base_subdir = self['project_name']
        base_dir = os.path.join(self.get_data_root_directory(**kwargs),
                                base_subdir)
        if container:
            user_project_dir = container.med.cfg.get('project_dir')
            directory = (os.path.join(base_dir, container.name)
                         if user_project_dir is None else
                         os.path.join(*ensure_collection(user_project_dir)))
        else:
            directory = base_dir
        if subdirs is not None:
            directory = os.path.join(directory, *subdirs)

        if makedirs:
            # Create directory if needed
            os.makedirs(directory, exist_ok=True)

        return directory

    def get_external_data_directory(
            self, container=None, makedirs=True, subdirs=None, **kwargs):
        """Get external data directory and create it if needed.

        :param container: Data source for which to get data directory.
        :param makedirs: Make directories if needed. Default is `True`.
        :param subdirs: List of additional subdirectories.
        :type container: :py:class:`.data_source.DataSourceBase`
        :type makedirs: bool
        :type subdirs: :py:class:`str` or :py:class:`list` or :py:class:`str`

        :returns: Directory path.
        :rtype: str
        """
        subdirs = ensure_collection(subdirs)
        base_subdir = 'extern'
        base_dir = os.path.join(self.get_data_root_directory(**kwargs),
                                base_subdir)
        if container:
            user_data_dir = container.cfg.get('data_dir')
            directory = (os.path.join(base_dir, container.name)
                         if user_data_dir is None else
                         os.path.join(*ensure_collection(user_data_dir)))
        else:
            directory = base_dir
        if subdirs is not None:
            directory = os.path.join(directory, *subdirs)

        if makedirs:
            # Create directory if needed
            os.makedirs(directory, exist_ok=True)

        return directory

    def get_plot_root_directory(self, **kwargs):
        """Get plot root directory from mediator configuration.

        :returns: Plot root-directory.
        :rtype: str
        """
        directory = self.get('plot_dir')
        directory = ('plot' if directory is None else
                     os.path.join(*ensure_collection(directory)))

        return directory

    def get_plot_directory(
            self, container=None, makedirs=True, subdirs=None, **kwargs):
        """Get project data directory and create it if needed.

        :param container: Container for which to get data directory.
        :param makedirs: Make directories if needed. Default is `True`.
        :param subdirs: List of additional subdirectories.
        :type container: :py:class:`.container.Container`
        :type makedirs:
        :type subdirs: :py:class:`str` or :py:class:`list` or :py:class:`str`

        :returns: Directory path.
        :rtype: str
        """
        subdirs = ensure_collection(subdirs)
        base_dir = os.path.join(self.get_plot_root_directory(**kwargs),
                                self['project_name'])

        directory = (os.path.join(base_dir, container.name)
                     if container is not None else base_dir)
        if subdirs is not None:
            directory = os.path.join(directory, *subdirs)

        if makedirs:
            # Create directory if needed
            os.makedirs(directory, exist_ok=True)

        return directory

    # Interface with cfg dictionary
    def __getitem__(self, key):
        """Get item from :py:attr:`cfg`."""
        return self.cfg[key]

    def __setitem__(self, key, value):
        """Set item in :py:attr:`cfg`."""
        self.cfg[key] = value

    def __contains__(self, key):
        """:py:attr:`cfg` contains."""
        return key in self.cfg

    def __delitem__(self, key):
        del self.cfg[key]

    def __iter__(self):
        """Iterate :py:attr:`cfg`."""
        return iter(self.cfg)

    def __len__(self, key, value):
        """Length of :py:obj:`cfg`."""
        return len(self.cfg)

    def __str__(self):
        """Return `str(self.cfg)`."""
        return str(self.cfg)

    def get(self, key, default=None):
        """Get item from :py:attr:`cfg`."""
        return self.cfg.get(key, default)


def import_module_from_cfg(cfg, name='', module_path=None):
    """Import actuator module from some container configuration.

    :param cfg: Configuration used to import module of container.
    :param name: Container name. Default is `''`.
    :param module_path: Module path. Default is `None`, in which
      case the module path is read from the `module_path` entry
      of the configuration.
    :type cfg: mapping
    :type name: str
    :type module_path: str

    :returns: Actuator module.
    :rtype: module
    """
    module_path = module_path or cfg.get('module_path')
    return (import_module(module_path, package=__package__)
            if module_path is not None else None)



"""data.gouv.fr API for French regions."""
import os
import requests
from ..geo import DefaultMaskAPI, DefaultMaskMaker, GEO_VARIABLE_NAME


class DataSource(DefaultMaskMaker, DefaultMaskAPI):
    #: Default source name.
    DEFAULT_SRC_NAME = 'datagouv_french_regions'

    #: Area: `'France'`.
    AREA = 'France'

    def __init__(self, med, name=None, cfg=None, **kwargs):
        """Naming constructor calling :py:class:`DefaultMaskMaker`'s
        constructor.

        :param med: Mediator.
        :param name: Data source name.
        :param cfg: Data source configuration. Default is `None`.
        :type med: :py:class:`.mediator.Mediator`
        :type name: str
        :type cfg: mapping
        """
        name = name or self.DEFAULT_SRC_NAME
        super(DataSource, self).__init__(med, name, cfg=cfg, **kwargs)

    def get_filename(self, **kwargs):
        """Get filename of geographical data for a given region.

        :returns: Shapefile filename.
        :rtype: str
        """
        return 'regions.shp'

    def download(self, **kwargs):
        """Download shapefile defining regions and
        store geographical data to :py:attr::`data` member.

        :returns: Names of downloaded variables.
        :rtype: :py:class:`set` of :py:class:`str`
        """
        # Make source dir
        src_dir = self.med.cfg.get_external_data_directory(self, **kwargs)

        # Get downloaded-file name
        filename_shp = self.get_filename(**kwargs)

        # Download
        root_url = '{}/fr/datasets/r/'.format(self.cfg['host'])
        self.log.info('Downloading shapefile from {}'.format(
            root_url))
        for filetype in self.cfg['filetypes']:
            filename = '{}.{}'.format(filename_shp[:-4], filetype)
            filepath = os.path.join(src_dir, filename)
            if filetype == 'prj':
                with open(filepath, 'w') as f:
                    # Replace Lambert II Carto with Lambert II Etendu
                    f.write(self.cfg['wkt'])
            else:
                url = '{}{}'.format(root_url, self.cfg[filetype])

                # Request and raise exception if needed
                response = requests.get(url)
                response.raise_for_status()

                # Write file
                with open(filepath, 'wb') as f:
                    for chunk in response:
                        f.write(chunk)

        return {GEO_VARIABLE_NAME}



"""Linear demand.actuator with heating and cooling days."""
import numpy as np
import pandas as pd
import xarray as xr
from sklearn import linear_model
from sklearn.model_selection import KFold, GridSearchCV
from sklearn.pipeline import Pipeline
from sklearn.exceptions import NotFittedError
from ..data_source import Composer
from ..actuator_base import ExtractorBase, EstimatorBase
from ..utils.learn import MultiInputRegressor, select_data


class Actuator(ExtractorBase, EstimatorBase):
    #: Default result name.
    DEFAULT_RESULT_NAME = 'demand'

    def __init__(self, result_mng, name, cfg=None, **kwargs):
        """Naming constructor.

        :param result_mng: Result manager of variable to estimate.
        :param name: Actuator name.
        :param cfg: Actuator configuration. Default is `None`.
        :type result_mng: :py:class:`ResultManager`
        :type name: str
        :type cfg: mapping
        """
        if result_mng.result_name != self.DEFAULT_RESULT_NAME:
            self.log.warning(
                'Result name {} given to constructor does not correspond '
                'to {} to be estimated by {}'.format(
                    result_mng.result_name, self.DEFAULT_RESULT_NAME,
                    self.name))

        #: Climate-variable name.
        self.climate_variable = 'surface_temperature'
        #: Calendar-variable name.
        self.calendar_variable = 'calendar'

        # Initialize extractor and transformer
        super(Actuator, self).__init__(
            result_mng=result_mng, name=name, cfg=cfg,
            variable_names=[self.climate_variable, self.calendar_variable],
            **kwargs)

        self.coef = {
            'daily_cycle_mean': None, 'regressor': None,
            'r2': None, 't_heat': None, 't_cool': None, 'alpha': None}

    def transform(self, data_src, stage=None, **kwargs):
        """Format temperature data from climate dataset.

        :param data_src: Input multiple data source containing climate
          and calendar data.
        :param stage: Modeling stage: `'fit'` or `'predict'`.
          May be required if features differ in the prediction stage.
        :type multidata_src: :py:class:`..data_source.MultiDataSource`
        :type stage: str

        :returns: Merged dataset.
        :rtype: :py:class:`xarray.Dataset`
        """
        # Get data sources
        clim_data_src = data_src.get_data_sources(self.climate_variable)
        cal_data_src = data_src.get_data_sources(self.calendar_variable)

        # Load climate data
        # Add regions domain cropping
        functions = [clim_data_src.crop_area]

        if hasattr(self.result_mng, 'modifier'):
            if (hasattr(self.result_mng.modifier.cfg, 'stage') and
                    (self.result_mng.modifier.cfg['stage'] != stage)):
                pass
            else:
                # Add modifier transformation
                functions.append(self.result_mng.modifier.transform)

        # Add get regional mean
        functions.append(clim_data_src.get_regional_mean)

        # Get temperature feature
        composer = Composer(*functions)

        ds_clim = clim_data_src.load(
            transform=composer, variable_names=self.climate_variable)

        # Convert from Kelvin to Celsius
        ds_clim[self.climate_variable] -= 273.15

        # Re-sample the climate data, if needed
        ds_clim = self._resample_data(ds_clim, reducer='mean', **kwargs)

        # Load calendar consistent with climate dataset
        ds_cal = cal_data_src.load(ds_clim.indexes['time'])

        # Return datasets merge
        return {**ds_clim, **ds_cal}

    def fit(self, data_src_in, data_src_out, **kwargs):
        """Learn statistical model of temperature-dependent demand
        and predict demand.

        :param data_src_in: Temperature and calendar dataset of training set.
        :param data_src_out: Demand dataset of training set.
        :type data_src_in: :py:class:`..data_source.DataSource`
        :type data_src_out: :py:class:`..data_source.DataSource`

        .. note::
          * This function uses grid search with k-fold cross validation to
            find the best heating/cooling temperature thresholds and
            regularization parameter.
          * These hyper-parameters are the same for all regions,
            so that the model has multiple outputs (one for each region)
            and the score for the cross valideation is given by the
            sum of the individual coefficients of determination for each region
            weighted by the fraction of the total variance explained by each
            region.
          * The feature matrix of the demand model is computed by calling the
            `fit` member function of the class
            :py:class:`DemandFeatureHeatCoolDays`, which itself calls
            the function :py:func:`_get_demand_feature_heat_cool_days`.

        """
        # Select component and result and reorder regions and coordinates
        da_clim = select_data(data_src_in, self.result_mng,
                              variable_name=self.climate_variable, **kwargs)
        da_cal = select_data(data_src_in, self.result_mng,
                             variable_name=self.calendar_variable, **kwargs)
        da_dem = select_data(data_src_out, self.result_mng, **kwargs)

        # Resample demand data, if needed
        da_dem = self._resample_data(da_dem, reducer='sum', **kwargs)

        # Assign variable coordinate
        da_clim = da_clim.expand_dims('variable').assign_coords(
            variable=[self.climate_variable])

        # Get common time index between demand and climate data
        da_clim, da_cal, da_dem = _select_time_slice(
            da_clim, da_cal, da_dem, **kwargs)

        # Get daily cycle mean
        if self.med.cfg['frequency'] == 'hour':
            self.coef['daily_cycle_mean'] = _get_daily_cycle(da_dem, da_cal)

        # Get raw design matrix
        des_mat_cycle = self._get_raw_design_matrix(da_clim, da_cal, **kwargs)

        # Cross-validation configuration
        if 'n_splits' in self.cfg:
            n_splits = self.cfg['n_splits']
        else:
            years = des_mat_cycle.indexes['time'].year
            n_splits = years[-1] - years[0] + 1
        # cv = TimeSeriesSplit(n_splits=self.cfg['n_splits'])
        cv = KFold(n_splits=n_splits)
        # scoring = 'neg_mean_squared_error'

        # Heating and cooling temperature grid
        t_heat_grid = np.arange(self.cfg['t_heat']['start'],
                                self.cfg['t_heat']['stop'],
                                self.cfg['t_heat']['step'])
        t_cool_grid = np.arange(self.cfg['t_cool']['start'],
                                self.cfg['t_cool']['stop'],
                                self.cfg['t_cool']['step'])

        # Estimator
        grid = {'feature__t_heat':  t_heat_grid,
                'feature__t_cool': t_cool_grid}
        params = {'fit_intercept': False, 'normalize': True}
        if 'max_iter' in self.cfg:
            params['max_iter'] = self.cfg['max_iter']
        if ((self.cfg['method'] == 'Ridge')
                | (self.cfg['method'] == 'Lasso')):
            alpha_grid = np.logspace(self.cfg['alpha']['start'],
                                     self.cfg['alpha']['stop'],
                                     self.cfg['alpha']['num'])
            grid.update({'regressor__estimator__alpha': alpha_grid})
            if self.cfg['method'] == 'Ridge':
                estimator = linear_model.Ridge(**params)
            elif self.cfg['method'] == 'Lasso':
                estimator = linear_model.Lasso(**params)
        elif self.cfg['method'] == 'BayesianRidge':
            estimator = linear_model.BayesianRidge(**params)

        # Make grid estimator
        daytype_index = np.unique(da_cal)
        feature = DemandFeatureHeatCoolDays(daytype_index=daytype_index)
        regressor = MultiInputRegressor(estimator)
        pipe = Pipeline(steps=[('feature', feature), ('regressor', regressor)])
        grid_cv = GridSearchCV(pipe, grid, cv=cv, verbose=0)

        # Fit model
        if not self.cfg.get('no_verbose'):
            self.log.info(
                'Fitting model by {}'.format(self.cfg['method']))
        grid_cv.fit(des_mat_cycle, da_dem.values)
        self.coef['regressor'], self.coef['r2'] = (
            grid_cv.best_estimator_, grid_cv.best_score_)

        # Get parameters of best model
        params = self.coef['regressor'].get_params()
        self.coef['t_heat'], self.coef['t_cool'] = (
            params['feature__t_heat'], params['feature__t_cool'])
        if self.cfg['method'] != 'BayesianRidge':
            self.coef['alpha'] = params['regressor__estimator__alpha']

        # Print parameters and scores
        if not self.cfg.get('no_verbose'):
            self.log.info(
                'Best overall score: {:.2f}'.format(self.coef['r2']))
            self.log.info('Heating temperature: {:.1f}'.format(
                self.coef['t_heat']))
            self.log.info('Cooling temperature: {:.1f}'.format(
                self.coef['t_cool']))
            if self.cfg['method'] != 'BayesianRidge':
                self.log.info('Regularization coefficient: {:.1f}'.format(
                    self.coef['alpha']))

    def predict(self, data_src_in, **kwargs):
        """Get regional demand prediction from fitted model.

        :param data_src_in: Input data source for prediction.
        :type data_src_in: :py:class:`..data_source.DataSource`

        :returns: Prediction dataset.
        :rtype: :py:class:`xarray.Dataset`
        """
        if self.coef['r2'] is None:
            raise NotFittedError(
                'This demand estimator instance has not been fitted yet')

        # Select component and result and reorder regions and coordinates
        da_clim = select_data(data_src_in, self.result_mng,
                              variable_name=self.climate_variable, **kwargs)
        da_cal = select_data(data_src_in, self.result_mng,
                             variable_name=self.calendar_variable, **kwargs)

        # Assign variable coordinate
        da_clim = da_clim.expand_dims('variable').assign_coords(
            variable=[self.climate_variable])

        # Get raw design matrix
        des_mat_cycle = self._get_raw_design_matrix(da_clim, da_cal, **kwargs)

        # Predict
        kwargs = ({'return_std': True} if self.cfg['method'] == 'BayesianRidge'
                  else {})
        pred = self.coef['regressor'].predict(des_mat_cycle, **kwargs)

        # Collect regional demand prediction
        ds_pred = xr.Dataset()
        coords = dict(des_mat_cycle.coords)
        coord_reg = ('region', des_mat_cycle.coords['region'])
        coord_time = ('time', coords['time'])
        if self.cfg['method'] == 'BayesianRidge':
            y_mean, y_std = pred
            # Add random perturbations to prediction drawn from
            # posterior distribution
            if self.med.cfg['frequency'] == 'hour':
                # Constant perturbations throughout the day
                y_pert_day = np.random.normal(loc=0., scale=y_std[::24])
                y_pert = np.empty(y_mean.shape)
                for ih in range(24):
                    y_pert[ih::24] = y_pert_day
            elif self.med.cfg['frequency'] == 'day':
                y_pert = np.random.normal(loc=0., scale=y_std)
            prediction = y_mean + y_pert

            # Add mean, standard deviation, alpha and lambda
            # y_mean = xr.DataArray(y_mean, coords=[coord_time, coord_reg],
            #                       name='demand_mean')
            y_std = xr.DataArray(y_std, coords=[coord_time, coord_reg],
                                 name='demand_std')
        else:
            prediction = pred

        # Add prediction
        prediction = xr.DataArray(prediction, coords=[coord_time, coord_reg])
        prediction.attrs['r2_total'] = self.coef['r2']
        prediction.attrs['t_heat'] = self.coef['t_heat']
        prediction.attrs['t_cool'] = self.coef['t_cool']
        if self.cfg['method'] != 'BayesianRidge':
            prediction.attrs['alpha'] = self.coef['alpha']

        # Add units
        if self.med.cfg['frequency'] == 'day':
            prediction.attrs['units'] = 'MWh/d'
        elif self.med.cfg['frequency'] == 'hour':
            prediction.attrs['units'] = 'MWh/h'
        ds_pred[self.result_mng.result_name] = prediction

        np.set_printoptions(precision=1)
        da_clim = des_mat_cycle.loc[{'variable': self.climate_variable}]
        val = prediction.mean('time').values * 365 / 1e6
        if not self.cfg.get('no_verbose'):
            self.log.info(
                'Total fitted demand mean (TWh/y): {}'.format(val))
            if self.cfg['method'] == 'BayesianRidge':
                y_std *= 365 / 1e6
                val = np.sqrt((y_std**2).mean('time').values)
                self.log.info(
                    'Total fitted demand std (TWh/y): {}'.format(val))
            # Get number of heating and cooling days
            n_heating_days = (da_clim < self.coef['t_heat']).mean(
                'time').values * 365
            n_cooling_days = (da_clim > self.coef['t_cool']).mean(
                'time').values * 365
            self.log.info('Number of heating days per year: {}'.format(
                n_heating_days))
            self.log.info('Number of cooling days per year: {}'.format(
                n_cooling_days))

        return ds_pred

    def get_extractor_postfix(self, **kwargs):
        """Get postfix corresponding to wind features.

        returns: Postfix.
        rtype: str
        """
        postfix = '{}_{}'.format(
            super(Actuator, self).get_extractor_postfix(**kwargs),
            self.med.cfg['frequency'])

        return postfix

    def get_estimator_postfix(self, **kwargs):
        """Get estimator postfix.

        returns: Postfix.
        rtype: str
        """
        return '{}_{}'.format(
            super(Actuator, self).get_estimator_postfix(**kwargs),
            self.cfg['method'])

    def _resample_data(self, data, reducer=None, **kwargs):
        """Resample dataset, if needed.

        :param data: Data.
        :param reducer: Name of method used to reduce after resampling.
        :type data: :py:class:`xarray.Dataset` or :py:class:`xarray.DataArray`
        :type reducer: str

        :returns: Data.
        :rtype: :py:class:`xarray.Dataset` or :py:class:`xarray.DataArray`
        """
        freq_data = data.indexes['time'].inferred_freq.upper()
        if ((freq_data in ['H', '1H']) and
                (self.med.cfg['frequency'] == 'day')):
            # Resample
            gp = data.resample(time='D')

            # Apply reduction
            meth = getattr(gp, reducer)
            data = meth('time', keep_attrs=True)
        elif ((freq_data in ['D', '1D']) and
              (self.med.cfg['frequency'] == 'hour')):
            end_time = data.indexes['time'][-1] + pd.Timedelta(23, unit='H')
            th = pd.date_range(
                start=data.indexes['time'][0], end=end_time, freq='H')
            data = data.reindex(time=th).ffill('time')

        return data

    def _get_raw_design_matrix(self, da_clim, da_cal, **kwargs):
        """Get raw design matrix.

        :param da_clim: Climate-data array.
        :param da_cal: Calendar-data array.
        :type da_clim: :py:class:`xarray.DataArray`
        :type da_cal: :py:class:`xarray.DataArray`

        :returns: Raw design matrix.
        :rtype: :py:class:`xarray.DataArray`
        """
        # Apply daily_cycle
        daily_cycle_series = _apply_daily_cycle(
            da_cal, self.coef.get('daily_cycle_mean'),
            self.med.cfg['frequency'])

        # Concatenate results and make sure dimensions in right order
        _, daily_cycle_series = xr.broadcast(
            da_clim, daily_cycle_series, exclude=['time', 'variable'])
        des_mat_cycle = xr.concat(
            [da_clim, daily_cycle_series], dim='variable')
        des_mat_cycle = des_mat_cycle.transpose('time', 'variable', 'region')

        return des_mat_cycle


def _get_demand_feature_heat_cool_days(
        des_mat, daytype_index, t_heat, t_cool):
    """Get feature matrix of demand model with as variables heating
    and cooling temperature ramps and as factors week-days types.

    :param des_mat: Array containing climatic variables, i.e.
        mean temperature(Celsius),
        and membership of days to types 'work', 'sat' and 'off'
        (possibly including a daily cycle).
    :param daytype_index: Daytype index.
    :param t_heat: Temperature threshold below which consumers
        turn on heating.
    :param t_cool: Temperature threshold above which consumers
        turn on air conditionning.
    :type des_mat: py:class:`xarray.DataArray`
    :type daytype_index: sequence
    :type t_heat: float
    :type t_cool: float

    :returns: Feature matrix.
    :rtype: :py:class:`xarray.DataArray`
    """
    temp_label = 'surface_temperature'
    # Get variable labels
    piece_name = ['one', 'heat', 'cool']
    variable_names = np.concatenate(
        [[['{}_{}'.format(v, day)] for v in piece_name]
         for day in daytype_index])[:, 0]

    # Build feature array
    coords = dict(des_mat.coords.items())  # Copy dictionary of coordinates
    # Update variable coordinate
    coords['variable'] = variable_names
    coords = [(dim, coords[dim]) for dim in des_mat.dims]
    shape = [len(coord[1]) for coord in coords]
    feature = xr.DataArray(np.zeros(shape), coords=coords,
                           name='demand_feature_heat_cool_days')

    # Select temperature
    z = des_mat.sel(variable=temp_label, drop=True)

    for day in daytype_index:
        # Day type mask
        id_day = des_mat.sel(variable=day, drop=True)

        # Intercept for each
        feature.loc[{'variable': 'one_{}'.format(day)}] = id_day

        # Heating temperature
        feature.loc[{'variable': 'heat_{}'.format(day)}] = (
            (t_heat - z) * np.heaviside(t_heat - z, 0.)) * id_day
        # (z < t_heat).astype(float))

        # Cooling temperature
        feature.loc[{'variable': 'cool_{}'.format(day)}] = (
            (z - t_cool) * np.heaviside(z - t_cool, 0.)) * id_day
        # (z > t_cool).astype(float)

    # Transpose with regions (outputs) as last dimension
    feature = feature.transpose('time', 'variable', 'region')

    return feature


class DemandFeatureHeatCoolDays():
    """Class given to a scikit-learn pipeline to extract feature
    of the piecewise-linear_model of demand as a function of
    temperature and type of days."""

    def __init__(self, daytype_index, t_heat=12., t_cool=18.):
        """Constructor.

        :param daytype_index: Day-type index.
        :param t_heat: Heating-temperature threshold. Default is `12.`.
        :param t_cool: Cooling-temperature threshold. Default is `12.`.
        :type daytype_index: sequence
        :type t_heat: float
        :type t_cool: float
        """
        #: Heating-temperature threshold.
        self.t_heat = t_heat

        #: Cooling-temperature threshold.
        self.t_cool = t_cool

        #: Day-type index.
        self.daytype_index = daytype_index

    def fit(self, des_mat, y=None):
        """Fit doing nothing."""
        return self

    def transform(self, des_mat):
        """Transform by calling :py:func:`_get_demand_feature_heat_cool_days`.

        :param des_mat: Design matrix.
        :type des_mat: :py:class:`xarray.DataArray`

        :returns: Feature matrix.
        :rtype: :py:class:`xarray.DataArray`
        """
        return _get_demand_feature_heat_cool_days(
            des_mat, self.daytype_index, self.t_heat, self.t_cool)

    def set_params(self, t_heat=None, t_cool=None, daytype_index=None):
        """Mimick `set_params` method from scikit-learn.

        .. todo:: Inheritance from scikit-learn was removed due to
          metaclass conflict when generating sphinx doc.
          Find a better solution.
        """
        # Set heating parameters
        if t_heat is not None:
            self.t_heat = t_heat
        # Set cooling parameters
        if t_cool is not None:
            self.t_cool = t_cool
        # Set daytype_index parameters
        if daytype_index is not None:
            self.daytype_index = daytype_index

    def get_params(self, **kwargs):
        """Mimick `set_params` method from scikit-learn.

        .. todo:: Inheritance from scikit-learn was removed due to
          metaclass conflict when generating sphinx doc.
          Find a better solution.
        """
        return {'t_heat': self.t_heat, 't_cool': self.t_cool,
                'daytype_index': self.daytype_index}


def _get_daily_cycle(da_dem, da_cal):
    """Build an array of complementary columns corresponding
    to each day type of the calendar.
    Non-zero values are given by a composite hourly cycle given by
    the average of demand over all samples of same day type and hour.

    :param da_dem: Demand array.
    :param da_cal: Calendar array.
    :type da_dem: :py:class:`xarray.DataArray`
    :type da_cal: :py:class:`xarray.DataArray`

    :returns: Array with each column corresponding to a day type.
    :rtype: :py:class:`xarray.DataArray`
    """
    # Group by day type, and hour if needed
    hours = da_cal.indexes['time'].hour
    hours.name = 'hour'
    da_dem.coords['daytype_hour'] = (
        'time', pd.MultiIndex.from_arrays([da_cal, hours]))
    gp_daily_cycle = da_dem.groupby('daytype_hour')

    # Get time-mean daily cycle per day type
    daily_cycle_mean = gp_daily_cycle.mean('time')

    return daily_cycle_mean


def _apply_daily_cycle(da_cal, daily_cycle_mean=None, frequency='day'):
    """Build an array of complementary columns corresponding
    to each day type of the calendar. If frequency is `'day'`,
    non-zero values are unitary, else, if frequency is `'hour'`,
    non-zero values are given by a composite hourly cycle given by
    the average of demand over all samples of same day type and hour.

    :param da_cal: Calendar array.
    :param daily_cycle_mean: Mean daily cycle of output data.
      Default is `None`.
    :param frequency: Sampling frequency as either `'day'` or `'hour'`.
       Default is `'day'`.
    :type da_cal: :py:class:`xarray.DataArray`
    :type daily_cycle_mean: :py:class:`xarray.DataArray`
    :type frequency: str

    :returns: An array with each column corresponding to a day type.
    :rtype: :py:class:`xarray.DataArray`
    """
    # Initialize cycle
    time = da_cal.indexes['time']
    daytype_index = np.unique(da_cal)
    coords = [('time', time), ('variable', daytype_index)]
    daily_cycle_series = xr.DataArray(
        np.zeros((len(time), len(daytype_index))), coords=coords)

    # Group by day type, and hour if needed
    if frequency == 'hour':
        hours = time.hour
        hours.name = 'hour'
        grouper = 'daytype_hour'
        group_index = pd.MultiIndex.from_arrays([da_cal, hours])
        daily_cycle_series.coords[grouper] = ('time', group_index)

        # Group calendar by daytype and hours
        gp_out = daily_cycle_series.groupby(grouper)

        # Fill cycle with composite
        daily_cycle_series = gp_out + daily_cycle_mean

        # Drop group variable
        daily_cycle_series = daily_cycle_series.drop(grouper)
    elif frequency == 'day':
        # Group calendar by daytype
        gp_out = daily_cycle_series.groupby(da_cal)
        for gp_key, gp_idx in gp_out.groups.items():
            daily_cycle_series.loc[{'variable': gp_key}][gp_idx] = 1

    return daily_cycle_series


def _select_time_slice(da_clim, da_cal, da_dem, **kwargs):
    """Select time slice from common index.

    :param da_clim: Climate data array.
    :param da_cal: Calendar data array.
    :param da_dem: Demand data array.
    :type da_clim: :py:class:`xarray.DataArray`
    :type da_cal: :py:class:`xarray.DataArray`
    :type da_dem: :py:class:`xarray.DataArray`

    :returns: Time-sliced data.
    :rtype: :py:class:`tuple` containing three
      :py:class:`xarray.DataArray`
    """
    common_index = da_dem.indexes['time']
    common_index = common_index.intersection(da_clim.indexes['time'])
    time_slice = slice(common_index[0], common_index[-1])

    # Select time slice
    da_clim = da_clim.sel(time=time_slice)
    da_cal = da_cal.sel(time=time_slice)
    da_dem = da_dem.sel(time=time_slice)

    return da_clim, da_cal, da_dem



"""ELIA (BE) API."""
import os
from requests import Session
from bs4 import BeautifulSoup
import pandas as pd
import xarray as xr
from ..data_source import DataSourceLoaderBase
from ..component import (DataSourceWithComponentsMixin,
                         parse_variable_component_args,
                         finalize_dataset)


class DataSource(DataSourceLoaderBase, DataSourceWithComponentsMixin):
    #: Default source name.
    DEFAULT_SRC_NAME = 'elia'

    #: Area: `'Belgium'`.
    AREA = 'Belgium'

    #: Default maximum fetch-trials.
    DEFAULT_MAX_FETCH_TRIALS = 50

    def __init__(self, med, name=None, cfg=None, **kwargs):
        """Naming constructor.

        :param med: Mediator.
        :param name: Data source name.
        :param cfg: Data source configuration. Default is `None`.
        :type med: :py:class:`.mediator.Mediator`
        :type name: str
        :type cfg: mapping
        """
        name = name or self.DEFAULT_SRC_NAME
        super(DataSource, self).__init__(med, name, cfg=cfg, **kwargs)

        #: Maximum fetch trials
        self.max_fetch_trials = (self.cfg.get('max_fetch_trials') or
                                 self.DEFAULT_MAX_FETCH_TRIALS)

        #: Requests session
        self._session = None

    @parse_variable_component_args
    def download(self, variable_component_names, **kwargs):
        """Download data.

        :param variable_component_names: Names of components to download per
          variable.
        :type variable_component_names: mapping from :py:class:`str`
          to :py:class:`set` of :py:class:`str`

        :returns: Names of downloaded components per variable.
        :rtype: mapping from :py:class:`str` to :py:class:`set` of
          :py:class:`str`

        .. note:: The data is downloaded from the
          `elia's REST API <https://publications.elia.be/Publications/Publications/>`.
        """
        # Start session
        self._session = Session()

        # Get region names
        _, src_region_place = self.med.geo_src.get_place_regions_for_source(
            self.name, self.AREA, **kwargs)

        # Get dates
        date_fmt = '%Y-%m-%d'
        dates = self._get_dates(**kwargs)

        # Define start and end dates
        svar = ', '.join(str(variable_name)
                         for variable_name in variable_component_names)
        if not self.cfg.get('no_verbose'):
            self.log.info('{} variables:'.format(svar))

        # Get region IDs
        region_ids = self._get_region_ids(**kwargs)

        # Loop over variables
        for variable_name, component_names in variable_component_names.items():
            src_variable_name = self.cfg['variable_names'][variable_name]
            if not self.cfg.get('no_verbose'):
                self.log.info('- {}:'.format(variable_name))

            # Loop over components
            for component_name in component_names:
                if not self.cfg.get('no_verbose'):
                    self.log.info('-- {}:'.format(component_name))

                for d, date in enumerate(dates[:-1]):
                    date_from = date.strftime(date_fmt)
                    date_to = dates[d + 1].strftime(date_fmt)
                    if not self.cfg.get('no_verbose'):
                        self.log.info('--- from {} to {}'.format(
                            date_from, date_to))

                    # Download all regions
                    for region_name, place_name in src_region_place.items():
                        source_id = region_ids[region_name]
                        if not self.cfg.get('no_verbose'):
                            self.log.info('---- {}'.format(region_name))

                        # Get items from REST API with BeautifulSoup
                        items = self._download_request(
                            component_name, date_from, date_to, source_id,
                            **kwargs)

                        if len(items) > 0:
                            # Define time index removing UTC timezone
                            time_list = items.find_all('DateTime')
                            time = pd.DatetimeIndex(
                                (v.text for v in time_list))
                            time.name = 'time'
                            time = time.tz_convert(None)

                            # Get series
                            cf_list = items.find_all(src_variable_name)
                            fact = self.cfg['unit_conversions'][
                                component_name][variable_name]
                            s = pd.Series(
                                (float(v.text) * fact if v.text else None
                                 for v in cf_list), index=time)
                        else:
                            s = pd.DataFrame()

                        # Save locally
                        filepath = self._get_download_filepath(
                            variable_name, component_name, region_name,
                            date_from, date_to, **kwargs)
                        s.to_csv(filepath)

        # Return names of downloaded variables
        return variable_component_names

    @parse_variable_component_args
    @finalize_dataset
    def load(self, variable_component_names, **kwargs):
        """Load data.

        :param variable_component_names: Names of components to load per
          variable.
        :type variable_component_names: mapping from :py:class:`str`
          to collection

        :returns: Time series for each variable and component.
        :rtype: dict
        """
        # Get region names
        _, src_region_place = self.med.geo_src.get_place_regions_for_source(
            self.name, self.AREA, **kwargs)

        # Get dates
        dates = self._get_dates(**kwargs)
        date_fmt = '%Y-%m-%d'
        start = pd.Timestamp(self.cfg['start_date'], tz='utc')
        end = pd.Timestamp(self.cfg['end_date'], tz='utc')

        # Loop over datasets
        ds = {}
        for variable_name, component_names in variable_component_names.items():
            if not self.cfg.get('no_verbose'):
                self.log.info('- {}:'.format(variable_name))
            # Loop over components
            da = None
            for component_name in component_names:
                if not self.cfg.get('no_verbose'):
                    self.log.info('-- {}:'.format(component_name))

                df = pd.DataFrame()
                for d, date in enumerate(dates[:-1]):
                    date_from = date.strftime(date_fmt)
                    date_to = dates[d + 1].strftime(date_fmt)
                    if not self.cfg.get('no_verbose'):
                        self.log.info('--- from {} to {}'.format(
                            date_from, date_to))

                    df_date = pd.DataFrame(columns=src_region_place.values(),
                                           dtype=float)
                    for region_name, place_name in src_region_place.items():
                        # Read downloaded data
                        filepath = self._get_download_filepath(
                            variable_name, component_name, region_name,
                            date_from, date_to, **kwargs)
                        s = pd.read_csv(
                            filepath, index_col=0, parse_dates=True).iloc[:, 0]

                        # Add region data to place
                        if df_date[place_name].any():
                            df_date[place_name] += s
                        else:
                            df_date[place_name] = s

                    # Select regions
                    df_date = df_date[src_region_place]

                    # Concatenate dates
                    df = pd.concat([df, df_date], axis='index')

                # Select time slice to remove extra hours from tz conversion
                df = df.loc[start:end]

                # Convert to DataArray
                da_comp = xr.DataArray(
                    df, dims=('time', 'region')).expand_dims(
                        'component').assign_coords(component=[component_name])

                # Concatenate
                da = (da_comp if da is None else
                      xr.concat([da, da_comp], dim='component'))

            # Convert capacity factor to hourly data
            da = da.resample(time='H').mean('time')

            # Add variable to dataset
            ds[variable_name] = da

        return ds

    def _get_url(self, component_name, date_from, date_to, source_id,
                 **kwargs):
        """Get URL.

        :param component_name: Component name.
        :param date_from: Date from which to start.
        :param date_to: Date to which to finish.
        :param source_id: Source ID.
        :type component_name: str
        :type date_from: str
        :type date_to: str
        :type source_id: str

        :returns: URL.
        :rtype: str
        """
        category = self.cfg['categories'][component_name]
        method = self.cfg['methods'][component_name]
        url_category = '{}/{}.{}.svc/{}'.format(
            self.cfg['host'], category, self.cfg['version'], method)
        try:
            url_options = '&'.join(
                '{}={}'.format(k, v) for k, v in self.cfg[
                    'options'][component_name].items())
        except AttributeError:
            url_options = ''
        url_cat_opt = '{}?{}'.format(url_category, url_options)

        if component_name == 'pv':
            url_date = '{}&dateFrom={}&dateTo={}'.format(
                url_cat_opt, date_from, date_to)
        elif component_name in ['wind', 'wind-onshore', 'wind-offshore']:
            url_date = '{}&beginDate={}&endDate={}'.format(
                url_cat_opt, date_from, date_to)

        # Add source ID to URL
        if component_name == 'pv':
            url = '{}&sourceId={}'.format(url_date, source_id)
        elif component_name in ['wind', 'wind-onshore', 'wind-offshore']:
            url = '{}&sourceIds=[{}]'.format(
                url_date, source_id)

        return url

    def _download_request(self, component_name, date_from, date_to, source_id,
                          **kwargs):
        """Get items from REST API with :py:class:`bs4.BeautifulSoup`.

        :param component_name: Component name.
        :param date_from: Date from which to start.
        :param date_to: Date to which to finish.
        :param source_id: Source ID.
        :type component_name: str
        :type date_from: str
        :type date_to: str
        :type source_id: str

        :returns items: Downloaded items.
        :rtype items: :py:class:`bs4.element.Tag`
        """
        # Get URL
        url = self._get_url(component_name, date_from, date_to, source_id,
                            **kwargs)

        tag_names = self.cfg['tags'][component_name]
        n_trials = 0
        while n_trials < self.max_fetch_trials:
            try:
                # Request and raise exception if needed
                response = self._session.get(url)
                response.raise_for_status()

                # Sparse XML
                soup = BeautifulSoup(response.text, 'xml')
                tag = getattr(soup, tag_names[0])
                items = getattr(tag, tag_names[1])

                # Everything went well -> leave trials loop
                break
            except (OSError, RuntimeError,
                    AttributeError, ValueError) as e:
                # Retry
                self.log.warning(
                    'Fetching trial {:d} failed: {}'.format(
                        n_trials + 1, str(e)))
                n_trials += 1
                continue

        # Verify that last trial succeeded
        if n_trials >= self.max_fetch_trials:
            # All trials failed
            self.log.critical('Fetching failed after {:d} '
                              'trials.'.format(n_trials))
            raise RuntimeError

        return items

    def _get_region_ids(self, **kwargs):
        """Get region IDs.

        :returns: Region-IDs mapping.
        :rtype: dict
        """
        method = 'GetRegions'
        category = 'SolarForecasting'
        url = '{}/{}.{}.svc/{}'.format(
            self.cfg['host'], category, self.cfg['version'], method)

        # Request and raise exception if needed
        response = self._session.get(url)
        response.raise_for_status()

        regions_list = response.json()
        region_ids = {v['Name']: v['SourceId'] for v in regions_list}

        return region_ids

    def _get_download_filepath(self, variable_name, component_name,
                               region_name, date_from, date_to, **kwargs):
        """Get downloaded dataset filepath.

        :param variable_name: Dataset name.
        :param component_name: Component name.
        :param region_name: Region name.
        :param date_from: Date from which to start.
        :param date_to: Date to which to end.
        :type variable_name: str
        :type component_name: str
        :type region_name: str
        :type date_from: str
        :type date_to: str

        :returns: Filepath
        :rtype: str
        """
        src_dir = self.med.cfg.get_external_data_directory(self)
        src_dir = os.path.join(src_dir, variable_name)
        os.makedirs(src_dir, exist_ok=True)
        filename = '{}_{}_{}_{}_{}-{}.csv'.format(
            self.name, variable_name, component_name, region_name,
            date_from, date_to)
        filepath = os.path.join(src_dir, filename)

        return filepath

    def _get_dates(self, **kwargs):
        """Define dates adding one day because request in Brussels time
        # (CET/CEST) at a daily frequency and timestamps returned in UTC.

        :returns: Dates.
        :rtype: :py:class:`pandas.DatetimeIndex`
        """
        freq = 'MS'
        start = pd.Timestamp(self.cfg['start_date'], tz='utc')
        end = pd.Timestamp(self.cfg['end_date'], tz='utc')
        dates = pd.date_range(start=start, end=end + pd.Timedelta(31, 'D'),
                              freq=freq)

        return dates

    def get_postfix(self, **kwargs):
        """Get data-source standard postfix.

        :returns: Postfix.
        :rtype: str
        """
        postfix = '_{}-{}'.format(
            self.cfg['start_date'], self.cfg['end_date'])

        return postfix



import os
import sys
from setuptools import setup, find_packages
sys.path.insert(0, os.getcwd())
import versioneer


# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


name = 'e4clim'
long_name = 'Energy for CLimate Integrated Model'
maintainer = 'alexis.tantet'
version = versioneer.get_version()
description = 'The {} for climate-aware renewable energy mixes'.format(
    long_name)
host = 'in2p3.fr'
home = 'https://gitlab.{}/{}/{}'.format(host, maintainer, name)
doc_url = 'https://{}.pages.{}/{}/'.format(maintainer, host, name)
license = 'GPLv3'
license_file = 'LICENSE.txt'

setup(
    name=name,
    long_name=long_name,
    version=version,
    cmdclass=versioneer.get_cmdclass(),
    author="{} developpers".format(long_name),
    author_email="{}@lmd.polytechnique.fr".format(maintainer),
    maintainer=maintainer,
    description=description,
    license=license,
    license_file=license_file,
    url=home,
    doc_url=doc_url,
    project_urls={"Documentation": doc_url},
    packages=find_packages(),
    long_description=read('README.md'),
    command_options={
        'build_sphinx': {}
    },
    install_requires=[
        'oyaml', 'numpy', 'scipy', 'pandas', 'numexpr',
        'xarray', 'dask', 'netcdf4', 'netCDF4', 'h5netcdf', 'cvxopt',
        'scikit-learn', 'matplotlib', 'python-dateutil', 'xlrd',
        'requests', 'geopandas', 'descartes',
        'uncertainties', 'rasterio', 'entsoe-py', 'requests_oauthlib'
    ],
    package_data={name: ['data/*', 'data/**/*', 'data/**/**/*',
                         'cfg/*', 'cfg/**/*']},
    entry_points={
        'console_scripts': [
            '{}-create-project = {}.bin.init:create_project'.format(
                name, name),
        ],
    }
)



"""MERRA-2 API."""
import os
from collections import OrderedDict
import requests
import netCDF4 as nc
import numpy as np
import pandas as pd
import xarray as xr
from ..grid import GriddedDataSourceBase
from ..container import ensure_collection


class DataSource(GriddedDataSourceBase):
    #: Default source name.
    DEFAULT_SRC_NAME = 'merra2'

    #: Frequency code.
    FREQ_CODES = {'hour': '1', 'day': 'D', 'month': 'M'}

    #: File sampling.
    FILE_SAMPLING = {'hour': 'D', 'day': 'D', 'month': 'M'}

    #: Frequency-dependent directory.
    DIR_FREQ = {'hour': '', 'day': '', 'month': '_MONTHLY'}

    #: Time-description code.
    TIME_DESCRIPTION_CODES = {
        'cnst': 'C', 'inst': 'I', 'stat': 'S', 'tavg': 'T'}

    #: Default maximum fetch-trials.
    DEFAULT_MAX_FETCH_TRIALS = 50

    def __init__(self, med, name=None, cfg=None, **kwargs):
        """Naming constructor.

        :param med: Mediator.
        :param name: Data source name.
        :param cfg: Data source configuration. Default is `None`.
        :type med: :py:class:`.mediator.Mediator`
        :type name: str
        :type cfg: mapping
        """
        name = name or self.DEFAULT_SRC_NAME
        super(DataSource, self).__init__(med, name, cfg=cfg, **kwargs)

        # If lat_range and lon_range are not given, they will
        # be computed from mask regions total bounds when needed
        self.lat_range = self.cfg.get('lat_range')
        self.lon_range = self.cfg.get('lon_range')

        # Time range
        self.time_range = self.cfg.get('time_range')
        if self.time_range is None:
            if ((self.cfg['time_description'] == 'stat')
                    or (self.cfg['frequency'] == 'month')):
                # Only one sample per day
                self.time_range_str = '[0:0]'
            elif self.cfg['time_description'] == 'tavg':
                # Default: get 24 hours of day
                self.time_range_str = '[0:23]'
        else:
            # User-defined time range
            self.time_range_str = '[{}:{}]'.format(
                self.time_range[0], self.time_range[1])

        # Server parent directory
        host_dir = (self.DEFAULT_SRC_NAME.upper()
                    + self.DIR_FREQ[self.cfg['frequency']])
        self.srv_parent_dir = '{}/opendap/hyrax/{}/{}{}{}{}{}'.format(
            self.cfg['host'], host_dir, self.cfg['data_name'],
            self.TIME_DESCRIPTION_CODES[self.cfg['time_description']],
            self.FREQ_CODES[self.cfg['frequency']],
            self.cfg['horizontal_resolution'],
            self.cfg['vertical_location'].upper())

        # Maximum fetch trials
        self.max_fetch_trials = (self.cfg.get('max_fetch_trials') or
                                 self.DEFAULT_MAX_FETCH_TRIALS)

    def download(self, variable_names=None, **kwargs):
        """Download merra2 data and save it to disk.

        :param variable_names: Names of variables to download.
          Default is `None`, in which case all variables in
          :py:attr:`variable_names` are downloaded.
        :type variable_names: (collection of) :py:class:`str`

        :returns: Names of downloaded variables.
        :rtype: :py:class:`set` of :py:class:`str`
        """
        # Get variable names
        variable_names = ensure_collection(
            variable_names, set) or self.variable_names

        # Loop over days
        date_range = pd.date_range(
            start=self.cfg['start_date'], end=self.cfg['end_date'],
            freq=self.FILE_SAMPLING[self.cfg['frequency']], closed='left')
        for date in date_range:
            for variable_name in variable_names:
                # Download data for date and variable
                self._download_date_variable(
                    date, variable_name, download=True, **kwargs)

        return variable_names

    def load(self, transform=None, variable_names=None, **kwargs):
        """Collect all required variables from the MERRA-2 re-analysis.

        :param transform: A function or a composee of functions
          to apply to the datasets.
          These functions should take as arguments a dataset and a
          :py:class:`.data_source.DataSourceBase` data source.
        :param variable_names: Names of variables to download.
          Default is `None`, in which case all variables in
          :py:attr:`variable_names` are downloaded.
        :type transform: :py:class:`func` or
          :py:class:`.data_source.Composer`
        :type variable_names: (collection of) :py:class:`str`

        :returns: A dataset collecting all variables and periods.
        :rtype: :py:class:`xarray.Dataset`
        """
        # Get variable names
        variable_names = ensure_collection(
            variable_names, set) or self.variable_names

        # Loop over days
        date_range = pd.date_range(
            start=self.cfg['start_date'], end=self.cfg['end_date'],
            freq=self.FILE_SAMPLING[self.cfg['frequency']], closed='left')
        ds = None
        for date in date_range:
            # Collect all variables from all groups
            ds_per = xr.Dataset()
            if not self.cfg.get('no_verbose'):
                self.log.info('Reading data for {}'.format(
                    date.date()))
            for variable_name in variable_names:
                src_variable_name = self.cfg['variable_names'].get(
                    variable_name)

                # Read data for date and variable
                ds_gp = self._download_date_variable(
                    date, variable_name, download=False, **kwargs)
                if ds_gp is None:
                    continue

                # Add height
                self._add_height(ds_gp)

                # Rename variable
                ds_gp = ds_gp.rename(**{src_variable_name: variable_name})

                # Merge group
                ds_per = ds_per.merge(ds_gp)
                ds_gp.close()

            # Remove conflicting attributes to avoid
            # AttributeError: NetCDF: String match to name in use
            try:
                del ds_per.time.attrs['CLASS']
                del ds_per.time.attrs['NAME']
            except KeyError:
                pass

            # Apply functions to the dataset if given
            if transform:
                kwargs.update({'ds': ds_per, 'data_src': self})
                ds_per = transform(**kwargs)

            # Create or add period
            ds = ds_per if ds is None else xr.concat([ds, ds_per], dim='time')

        # Change hour interval representation convention from center to left
        t = ds.indexes['time']
        index_parts = {'year': t.year, 'month': t.month, 'day': t.day}
        if self.cfg['frequency'] == 'hour':
            index_parts['hour'] = t.hour
        new_index = pd.to_datetime(index_parts)
        ds = ds.reindex(time=new_index, method='bfill')
        ds.time.encoding['units'] = "hours since 1980-01-01T00:00:00"

        return ds

    def get_postfix(self, start_date=None, end_date=None, **kwargs):
        """Get postfix for the MERRA-2 data.

        :param start_date: Simulation start-date. Default is `None`, in which
          case the first date of `self.cfg['streams']['start_dates']` is used.
        :param end_date: Simulation end-date. Default is `None`, in which case
          the first date of `self.cfg['streams']['end_dates']` is used.
        :type start_date: str
        :type end_date: str

        :returns: Postfix.
        :rtype: str
        """
        start_date = start_date or self.cfg['start_date']
        end_date = end_date or self.cfg['end_date']

        # Get paths
        postfix = '_{}_{}-{}'.format(self.cfg['frequency'],
                                     start_date, end_date)

        return postfix

    def get_index_lon(self, lon):
        """Get grid index corresponding to longitude.

        :param lon: Longitude.
        :type lon: :py:class:`float`, array_like

        :returns: Longitude grid-index.
        :rtype: :py:class:`int`, :py:class:`numpy.array`
        """
        ilon = (lon + 180.) / self.cfg['delta_lon']
        try:
            ilon = int(ilon + 0.1)
        except TypeError:
            ilon = (np.array(ilon) + 0.1).astype(int)

        return ilon

    def get_index_lat(self, lat):
        """Get grid index corresponding to latitude.

        :param lat: Latitude.
        :type lat: :py:class:`float`, array_like

        :returns: Latitude grid-index.
        :rtype: :py:class:`int`, :py:class:`numpy.array`
        """
        ilat = (lat + 90.) / self.cfg['delta_lat']
        try:
            ilat = int(ilat + 0.1)
        except TypeError:
            ilat = (np.array(ilat) + 0.1).astype(int)

        return ilat

    def get_lat_lon_range_from_geo_bounds(self):
        """Get latitude and longitude ranges from bounds of geographic
        data source."""
        # Get total bounds in geoid system
        lon_min, lat_min, lon_max, lat_max = (
            self.med.geo_src.get_total_bounds(epsg=4326))

        # Set ranges
        self.lat_range = [lat_min, lat_max]
        self.lon_range = [lon_min, lon_max]

    def get_lat_lon_range_str(self):
        """Convert lat/lon range to indices string."""
        lat_min = self.lat_range[0]
        lat_max = self.lat_range[1]
        lon_min = self.lon_range[0]
        lon_max = self.lon_range[1]

        # Convert bounds to grid indices
        ilon_min = self.get_index_lon(lon_min)
        ilon_max = self.get_index_lon(lon_max) + 1
        ilat_min = self.get_index_lat(lat_min)
        ilat_max = self.get_index_lat(lat_max) + 1

        # Get grid string
        self.lon_range_str = '[{:d}:{:d}]'.format(ilon_min, ilon_max)
        self.lat_range_str = '[{:d}:{:d}]'.format(ilat_min, ilat_max)

    def get_url_file_dir_file(self, variable_name=None, date=None):
        """Get variable URL, directory and filename for MERRA-2 data.

        :param variable_name: Variable name. Default is `None`.
        :param date: Date for which to return paths.
          If `None`, use `'start_date'` value of data-source configuration.
          Default is `None`.
        :type variable_name: str

        returns: Source URL, directory and filename.
        :rtype: :py:class:`tuple` of :py:class:`str`
        """
        if (not self.lat_range) or (not self.lon_range):
            # Get lat/lon range from bounds of geographic data
            self.get_lat_lon_range_from_geo_bounds()

        # Convert lat/lon range to indices string
        self.get_lat_lon_range_str()

        # Define grid string
        domain = '{}{}{}'.format(self.time_range_str, self.lat_range_str,
                                 self.lon_range_str)
        short_time_range_str = ('' if self.cfg['space'] == '2d'
                                else self.time_range_str)
        grid_list = 'time{},lat{},lon{}'.format(
            short_time_range_str, self.lat_range_str, self.lon_range_str)

        # Get data-source directory
        src_dir = self.med.cfg.get_external_data_directory(self)

        # Get given date or start date
        date = pd.Timestamp(date or self.cfg['start_date'])

        # Get runid of stream containg date
        for (stream, cfg_stream) in self.cfg['streams'].items():
            ssd = pd.Timestamp(cfg_stream['start_date'])
            sed = pd.Timestamp(cfg_stream['end_date'])
            if (date >= ssd) & (date < sed):
                break
        runid = '{}{}'.format(stream, self.cfg['version'])
        freq = '{}{}'.format(self.cfg['time_description'],
                             self.FREQ_CODES[self.cfg['frequency']])
        prefix0 = '{}_{}.{}_{}'.format(
            self.DEFAULT_SRC_NAME.upper(), runid, freq, self.cfg['space'])

        if variable_name is None:
            # Take first variable in a group
            group_name = None
            it = 0
            while group_name is None:
                variable_name = list(self.variable_names)[it]
                group_name = self.cfg['group_names'].get(variable_name)
                it += 1
        else:
            # Try to get group of variable
            group_name = self.cfg['group_names'].get(variable_name)
        src_variable_name = self.cfg['variable_names'].get(variable_name)

        # Make local directories
        file_dir = os.path.join(src_dir, group_name)
        os.makedirs(file_dir, exist_ok=True)

        # Frequency dependent variables
        date_dir = '{:04d}'.format(date.year)
        if self.cfg['frequency'] in ['hour', 'day']:
            date_dir += '/{:02d}'.format(date.month)
            date_file = date.strftime('%Y%m%d')
        else:
            date_file = date.strftime('%Y%m')

        HV = '{}{}'.format(self.cfg['horizontal_resolution'],
                           self.cfg['vertical_location'])
        prefix = '{}_{}_{}.{}.{}'.format(
            prefix0, group_name, HV, date_file, self.cfg['format'])
        postfix = '{}{},{}'.format(src_variable_name, domain, grid_list)
        filename = '{}.nc?{}'.format(prefix, postfix)

        srv_dir = '{}{}.{}'.format(
            self.srv_parent_dir, group_name.upper(), self.cfg['geos5_version'])
        url = '{}/{}/{}'.format(srv_dir, date_dir, filename)

        return url, file_dir, filename

    def get_grid(self, *args, **kwargs):
        """Get grid used to make masks.

        :returns: Dataset coordinates.
        :rtype: :py:class:`tuple` of pairs of :py:class:`str`
          and :py:class:`numpy.array`
        """
        # Get paths
        url, file_dir, filename = self.get_url_file_dir_file()

        # Read coords
        filepath = os.path.join(file_dir, filename)
        if not self.cfg.get('no_verbose'):
            self.log.info('Reading coordinates from {}'.format(filepath))
        with xr.open_dataset(filepath) as ds:
            coords = OrderedDict({
                'lat': ds['lat'].copy(deep=True).values,
                'lon': ds['lon'].copy(deep=True).values})

        return coords

    def get_grid_postfix(self, *args, **kwargs):
        """Grid postfix.

        :returns: Postfix.
        :rtype: str
        """
        postfix = '_{}{}'.format(self.cfg['horizontal_resolution'],
                                 self.cfg['vertical_location'])

        return postfix

    def _download_request(self, url):
        """Download data over HTTP - OPeNDAP.

        :param url: OPeNDAP URL.
        :type url: str

        :returns: Dataset.
        :rtype: :py:class:`xarray.Dataset`
        """
        # Request and raise exception if needed
        response = requests.get(url)
        response.raise_for_status()

        # Convert bytes to netCDF4.Dataset
        ds_nc = nc.Dataset(filename='dum.nc', memory=response.content)

        # Convert netCDF4.Dataset to xarray.Dataset
        ds = xr.open_dataset(xr.backends.NetCDF4DataStore(ds_nc))

        return ds

    def _download_date_variable(
            self, date, variable_name, download=False, **kwargs):
        """Download data for given date and variable.

        :param date: Date.
        :param variable_name: Variable name.
        :param download: Whether to direclty download data, or to try to
          read it from disk first. Default is `False`, in which case
          an attempt is made to read first.
        :type date: datetime
        :type variable_name: str
        :type download: bool
        """
        # Get group name for variable
        group_name = self.cfg['group_names'].get(variable_name)
        src_variable_name = self.cfg['variable_names'].get(variable_name)
        if (src_variable_name is None) or (group_name is None):
            self.log.warning(
                "{} variable not in 'variable_names' of {} "
                "configuration-file: skipping".format(
                    variable_name, self.name))
            return

        # Get paths
        url, file_dir, filename = self.get_url_file_dir_file(
            variable_name=variable_name, date=date)

        # Fetch data
        n_trials = 0
        while n_trials < self.max_fetch_trials:
            try:
                if (not download) and (n_trials == 0):
                    # Read previously downloaded data
                    filepath = os.path.join(file_dir, filename)
                    ds = xr.open_dataset(filepath)
                else:
                    # Try to fetch
                    if not self.cfg.get('no_verbose'):
                        self.log.info('Fetching {} group from {}'.format(
                            group_name, url))
                    ds = self._download_request(url)

                # Check quality
                if self.cfg.get('check_quality'):
                    _quality_check(ds)

                # Everything went well -> leave trials loop
                break
            except (OSError, RuntimeError,
                    AttributeError, ValueError) as e:
                # Retry
                self.log.warning(
                    'Fetching trial {:d} failed: {}'.format(
                        n_trials + 1, str(e)))
                n_trials += 1
                continue

        # Verify that last trial succeeded
        if n_trials >= self.max_fetch_trials:
            # All trials failed
            self.log.critical('Fetching failed after {:d} '
                              'trials.'.format(n_trials))
            raise RuntimeError
        elif download or (n_trials > 0):
            # Write data
            filepath = os.path.join(file_dir, filename)
            if not self.cfg.get('no_verbose'):
                self.log.info('Writing {} group to {}'.format(
                    group_name, filepath))
            ds.to_netcdf(filepath)

        return ds

    def _add_height(self, ds):
        """Add or rename `'height'` variable to dataset attributes.

        :param ds: Dataset.
        :type ds: :py:class:`xarray.Dataset`
        """
        for svn in ds.data_vars:
            var = ds[svn]
            if self.cfg['height'] not in var.attrs:
                # Add height as attribute
                if self.cfg['height'] in ds:
                    var.attrs['height'] = float(
                        ds[self.cfg['height']])
                elif self.cfg['height'] in ds.coords:
                    var.attrs['height'] = float(
                        ds.coords[self.cfg['height']])
                else:
                    # Get height from wind name or assume 2m height
                    var.attrs['height'] = (
                        float(svn[1:svn.find('M')])
                        if svn[0] in ['U', 'V'] else 2.)
            else:
                # Rename
                var.attrs['height'] = var.attrs[self.cfg['height']]


def _quality_check(ds):
    """Check integrity of downloaded dataset. If dataset is invalid,
    `ValueError` is raised.

    :param ds: Dataset to check.
    :type ds: :py:class:`xarray.Dataset`
    """
    # Time index
    if (ds.indexes['time'].year == 1970).any():
        raise ValueError

    # NaN and Infinite values
    for da in ds.values():
        if da.isnull().any() or (da > 1e10).any():
            raise ValueError
